# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Entry points into the Face layer of RPC Framework."""

from grpc.framework.common import cardinality
from grpc.framework.common import style
from grpc.framework.base import exceptions as _base_exceptions
from grpc.framework.base import interfaces as base_interfaces
from grpc.framework.face import _calls
from grpc.framework.face import _service
from grpc.framework.face import exceptions
from grpc.framework.face import interfaces


class _BaseServicer(base_interfaces.Servicer):

  def __init__(self, methods, multi_method):
    self._methods = methods
    self._multi_method = multi_method

  def service(self, name, context, output_consumer):
    method = self._methods.get(name, None)
    if method is not None:
      return method(output_consumer, context)
    elif self._multi_method is not None:
      try:
        return self._multi_method.service(name, output_consumer, context)
      except exceptions.NoSuchMethodError:
        raise _base_exceptions.NoSuchMethodError()
    else:
      raise _base_exceptions.NoSuchMethodError()


class _UnaryUnaryMultiCallable(interfaces.UnaryUnaryMultiCallable):

  def __init__(self, front, name):
    self._front = front
    self._name = name

  def __call__(self, request, timeout):
    return _calls.blocking_value_in_value_out(
        self._front, self._name, request, timeout, 'unused trace ID')

  def future(self, request, timeout):
    return _calls.future_value_in_value_out(
        self._front, self._name, request, timeout, 'unused trace ID')

  def event(self, request, response_callback, abortion_callback, timeout):
    return _calls.event_value_in_value_out(
        self._front, self._name, request, response_callback, abortion_callback,
        timeout, 'unused trace ID')


class _UnaryStreamMultiCallable(interfaces.UnaryStreamMultiCallable):

  def __init__(self, front, name):
    self._front = front
    self._name = name

  def __call__(self, request, timeout):
    return _calls.inline_value_in_stream_out(
        self._front, self._name, request, timeout, 'unused trace ID')

  def event(self, request, response_consumer, abortion_callback, timeout):
    return _calls.event_value_in_stream_out(
        self._front, self._name, request, response_consumer, abortion_callback,
        timeout, 'unused trace ID')


class _StreamUnaryMultiCallable(interfaces.StreamUnaryMultiCallable):

  def __init__(self, front, name, pool):
    self._front = front
    self._name = name
    self._pool = pool

  def __call__(self, request_iterator, timeout):
    return _calls.blocking_stream_in_value_out(
        self._front, self._name, request_iterator, timeout, 'unused trace ID')

  def future(self, request_iterator, timeout):
    return _calls.future_stream_in_value_out(
        self._front, self._name, request_iterator, timeout, 'unused trace ID',
        self._pool)

  def event(self, response_callback, abortion_callback, timeout):
    return _calls.event_stream_in_value_out(
        self._front, self._name, response_callback, abortion_callback, timeout,
        'unused trace ID')


class _StreamStreamMultiCallable(interfaces.StreamStreamMultiCallable):

  def __init__(self, front, name, pool):
    self._front = front
    self._name = name
    self._pool = pool

  def __call__(self, request_iterator, timeout):
    return _calls.inline_stream_in_stream_out(
        self._front, self._name, request_iterator, timeout, 'unused trace ID',
        self._pool)

  def event(self, response_consumer, abortion_callback, timeout):
    return _calls.event_stream_in_stream_out(
        self._front, self._name, response_consumer, abortion_callback, timeout,
        'unused trace ID')


class _GenericStub(interfaces.GenericStub):
  """An interfaces.GenericStub implementation."""

  def __init__(self, front, pool):
    self._front = front
    self._pool = pool

  def blocking_value_in_value_out(self, name, request, timeout):
    return _calls.blocking_value_in_value_out(
        self._front, name, request, timeout, 'unused trace ID')

  def future_value_in_value_out(self, name, request, timeout):
    return _calls.future_value_in_value_out(
        self._front, name, request, timeout, 'unused trace ID')

  def inline_value_in_stream_out(self, name, request, timeout):
    return _calls.inline_value_in_stream_out(
        self._front, name, request, timeout, 'unused trace ID')

  def blocking_stream_in_value_out(self, name, request_iterator, timeout):
    return _calls.blocking_stream_in_value_out(
        self._front, name, request_iterator, timeout, 'unused trace ID')

  def future_stream_in_value_out(self, name, request_iterator, timeout):
    return _calls.future_stream_in_value_out(
        self._front, name, request_iterator, timeout, 'unused trace ID',
        self._pool)

  def inline_stream_in_stream_out(self, name, request_iterator, timeout):
    return _calls.inline_stream_in_stream_out(
        self._front, name, request_iterator, timeout, 'unused trace ID',
        self._pool)

  def event_value_in_value_out(
      self, name, request, response_callback, abortion_callback, timeout):
    return _calls.event_value_in_value_out(
        self._front, name, request, response_callback, abortion_callback,
        timeout, 'unused trace ID')

  def event_value_in_stream_out(
      self, name, request, response_consumer, abortion_callback, timeout):
    return _calls.event_value_in_stream_out(
        self._front, name, request, response_consumer, abortion_callback,
        timeout, 'unused trace ID')

  def event_stream_in_value_out(
      self, name, response_callback, abortion_callback, timeout):
    return _calls.event_stream_in_value_out(
        self._front, name, response_callback, abortion_callback, timeout,
        'unused trace ID')

  def event_stream_in_stream_out(
      self, name, response_consumer, abortion_callback, timeout):
    return _calls.event_stream_in_stream_out(
        self._front, name, response_consumer, abortion_callback, timeout,
        'unused trace ID')

  def unary_unary_multi_callable(self, name):
    return _UnaryUnaryMultiCallable(self._front, name)

  def unary_stream_multi_callable(self, name):
    return _UnaryStreamMultiCallable(self._front, name)

  def stream_unary_multi_callable(self, name):
    return _StreamUnaryMultiCallable(self._front, name, self._pool)

  def stream_stream_multi_callable(self, name):
    return _StreamStreamMultiCallable(self._front, name, self._pool)


class _DynamicStub(interfaces.DynamicStub):
  """An interfaces.DynamicStub implementation."""

  def __init__(self, cardinalities, front, pool):
    self._cardinalities = cardinalities
    self._front = front
    self._pool = pool

  def __getattr__(self, attr):
    method_cardinality = self._cardinalities.get(attr)
    if method_cardinality is cardinality.Cardinality.UNARY_UNARY:
      return _UnaryUnaryMultiCallable(self._front, attr)
    elif method_cardinality is cardinality.Cardinality.UNARY_STREAM:
      return _UnaryStreamMultiCallable(self._front, attr)
    elif method_cardinality is cardinality.Cardinality.STREAM_UNARY:
      return _StreamUnaryMultiCallable(self._front, attr, self._pool)
    elif method_cardinality is cardinality.Cardinality.STREAM_STREAM:
      return _StreamStreamMultiCallable(self._front, attr, self._pool)
    else:
      raise AttributeError('_DynamicStub object has no attribute "%s"!' % attr)


def _adapt_method_implementations(method_implementations, pool):
  adapted_implementations = {}
  for name, method_implementation in method_implementations.iteritems():
    if method_implementation.style is style.Service.INLINE:
      if method_implementation.cardinality is cardinality.Cardinality.UNARY_UNARY:
        adapted_implementations[name] = _service.adapt_inline_value_in_value_out(
            method_implementation.unary_unary_inline)
      elif method_implementation.cardinality is cardinality.Cardinality.UNARY_STREAM:
        adapted_implementations[name] = _service.adapt_inline_value_in_stream_out(
            method_implementation.unary_stream_inline)
      elif method_implementation.cardinality is cardinality.Cardinality.STREAM_UNARY:
        adapted_implementations[name] = _service.adapt_inline_stream_in_value_out(
            method_implementation.stream_unary_inline, pool)
      elif method_implementation.cardinality is cardinality.Cardinality.STREAM_STREAM:
        adapted_implementations[name] = _service.adapt_inline_stream_in_stream_out(
            method_implementation.stream_stream_inline, pool)
    elif method_implementation.style is style.Service.EVENT:
      if method_implementation.cardinality is cardinality.Cardinality.UNARY_UNARY:
        adapted_implementations[name] = _service.adapt_event_value_in_value_out(
            method_implementation.unary_unary_event)
      elif method_implementation.cardinality is cardinality.Cardinality.UNARY_STREAM:
        adapted_implementations[name] = _service.adapt_event_value_in_stream_out(
            method_implementation.unary_stream_event)
      elif method_implementation.cardinality is cardinality.Cardinality.STREAM_UNARY:
        adapted_implementations[name] = _service.adapt_event_stream_in_value_out(
            method_implementation.stream_unary_event)
      elif method_implementation.cardinality is cardinality.Cardinality.STREAM_STREAM:
        adapted_implementations[name] = _service.adapt_event_stream_in_stream_out(
            method_implementation.stream_stream_event)
  return adapted_implementations


def servicer(pool, method_implementations, multi_method_implementation):
  """Creates a base_interfaces.Servicer.

  It is guaranteed that any passed interfaces.MultiMethodImplementation will
  only be called to service an RPC if there is no
  interfaces.MethodImplementation for the RPC method in the passed
  method_implementations dictionary.

  Args:
    pool: A thread pool.
    method_implementations: A dictionary from RPC method name to
      interfaces.MethodImplementation object to be used to service the named
      RPC method.
    multi_method_implementation: An interfaces.MultiMethodImplementation to be
      used to service any RPCs not serviced by the
      interfaces.MethodImplementations given in the method_implementations
      dictionary, or None.

  Returns:
    A base_interfaces.Servicer that services RPCs via the given implementations.
  """
  adapted_implementations = _adapt_method_implementations(
      method_implementations, pool)
  return _BaseServicer(adapted_implementations, multi_method_implementation)


def generic_stub(front, pool):
  """Creates an interfaces.GenericStub.

  Args:
    front: A base_interfaces.Front.
    pool: A futures.ThreadPoolExecutor.

  Returns:
    An interfaces.GenericStub that performs RPCs via the given
      base_interfaces.Front.
  """
  return _GenericStub(front, pool)


def dynamic_stub(cardinalities, front, pool, prefix):
  """Creates an interfaces.DynamicStub.

  Args:
    cardinalities: A dict from RPC method name to cardinality.Cardinality
      value identifying the cardinality of every RPC method to be supported by
      the created interfaces.DynamicStub.
    front: A base_interfaces.Front.
    pool: A futures.ThreadPoolExecutor.
    prefix: A string to prepend when mapping requested attribute name to RPC
      method name during attribute access on the created
      interfaces.DynamicStub.

  Returns:
    An interfaces.DynamicStub that performs RPCs via the given
      base_interfaces.Front.
  """
  return _DynamicStub(cardinalities, front, pool)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test of the GRPC-backed ForeLink and RearLink."""

import threading
import unittest

from grpc._adapter import _proto_scenarios
from grpc._adapter import _test_links
from grpc._adapter import fore
from grpc._adapter import rear
from grpc.framework.base import interfaces
from grpc.framework.foundation import logging_pool

_IDENTITY = lambda x: x
_TIMEOUT = 32


# TODO(nathaniel): End-to-end metadata testing.
def _transform_metadata(unused_metadata):
  return (
      ('one unused key', 'one unused value'),
      ('another unused key', 'another unused value'),
)


class RoundTripTest(unittest.TestCase):

  def setUp(self):
    self.fore_link_pool = logging_pool.pool(8)
    self.rear_link_pool = logging_pool.pool(8)

  def tearDown(self):
    self.rear_link_pool.shutdown(wait=True)
    self.fore_link_pool.shutdown(wait=True)

  def testZeroMessageRoundTrip(self):
    test_operation_id = object()
    test_method = 'test method'
    test_fore_link = _test_links.ForeLink(None, None)
    def rear_action(front_to_back_ticket, fore_link):
      if front_to_back_ticket.kind in (
          interfaces.FrontToBackTicket.Kind.COMPLETION,
          interfaces.FrontToBackTicket.Kind.ENTIRE):
        back_to_front_ticket = interfaces.BackToFrontTicket(
            front_to_back_ticket.operation_id, 0,
            interfaces.BackToFrontTicket.Kind.COMPLETION, None)
        fore_link.accept_back_to_front_ticket(back_to_front_ticket)
    test_rear_link = _test_links.RearLink(rear_action, None)

    fore_link = fore.ForeLink(
        self.fore_link_pool, {test_method: None}, {test_method: None}, None, ())
    fore_link.join_rear_link(test_rear_link)
    test_rear_link.join_fore_link(fore_link)
    fore_link.start()
    port = fore_link.port()

    rear_link = rear.RearLink(
        'localhost', port, self.rear_link_pool, {test_method: None},
        {test_method: None}, False, None, None, None,
        metadata_transformer=_transform_metadata)
    rear_link.join_fore_link(test_fore_link)
    test_fore_link.join_rear_link(rear_link)
    rear_link.start()

    front_to_back_ticket = interfaces.FrontToBackTicket(
        test_operation_id, 0, interfaces.FrontToBackTicket.Kind.ENTIRE,
        test_method, interfaces.ServicedSubscription.Kind.FULL, None, None,
        _TIMEOUT)
    rear_link.accept_front_to_back_ticket(front_to_back_ticket)

    with test_fore_link.condition:
      while (not test_fore_link.tickets or
             test_fore_link.tickets[-1].kind is
                 interfaces.BackToFrontTicket.Kind.CONTINUATION):
        test_fore_link.condition.wait()

    rear_link.stop()
    fore_link.stop()

    with test_fore_link.condition:
      self.assertIs(
          test_fore_link.tickets[-1].kind,
          interfaces.BackToFrontTicket.Kind.COMPLETION)

  def testEntireRoundTrip(self):
    test_operation_id = object()
    test_method = 'test method'
    test_front_to_back_datum = b'\x07'
    test_back_to_front_datum = b'\x08'
    test_fore_link = _test_links.ForeLink(None, None)
    rear_sequence_number = [0]
    def rear_action(front_to_back_ticket, fore_link):
      if front_to_back_ticket.payload is None:
        payload = None
      else:
        payload = test_back_to_front_datum
      terminal = front_to_back_ticket.kind in (
          interfaces.FrontToBackTicket.Kind.COMPLETION,
          interfaces.FrontToBackTicket.Kind.ENTIRE)
      if payload is not None or terminal:
        if terminal:
          kind = interfaces.BackToFrontTicket.Kind.COMPLETION
        else:
          kind = interfaces.BackToFrontTicket.Kind.CONTINUATION
        back_to_front_ticket = interfaces.BackToFrontTicket(
            front_to_back_ticket.operation_id, rear_sequence_number[0], kind,
            payload)
        rear_sequence_number[0] += 1
        fore_link.accept_back_to_front_ticket(back_to_front_ticket)
    test_rear_link = _test_links.RearLink(rear_action, None)

    fore_link = fore.ForeLink(
        self.fore_link_pool, {test_method: _IDENTITY},
        {test_method: _IDENTITY}, None, ())
    fore_link.join_rear_link(test_rear_link)
    test_rear_link.join_fore_link(fore_link)
    fore_link.start()
    port = fore_link.port()

    rear_link = rear.RearLink(
        'localhost', port, self.rear_link_pool, {test_method: _IDENTITY},
        {test_method: _IDENTITY}, False, None, None, None)
    rear_link.join_fore_link(test_fore_link)
    test_fore_link.join_rear_link(rear_link)
    rear_link.start()

    front_to_back_ticket = interfaces.FrontToBackTicket(
        test_operation_id, 0, interfaces.FrontToBackTicket.Kind.ENTIRE,
        test_method, interfaces.ServicedSubscription.Kind.FULL, None,
        test_front_to_back_datum, _TIMEOUT)
    rear_link.accept_front_to_back_ticket(front_to_back_ticket)

    with test_fore_link.condition:
      while (not test_fore_link.tickets or
             test_fore_link.tickets[-1].kind is not
                 interfaces.BackToFrontTicket.Kind.COMPLETION):
        test_fore_link.condition.wait()

    rear_link.stop()
    fore_link.stop()

    with test_rear_link.condition:
      front_to_back_payloads = tuple(
          ticket.payload for ticket in test_rear_link.tickets
          if ticket.payload is not None)
    with test_fore_link.condition:
      back_to_front_payloads = tuple(
          ticket.payload for ticket in test_fore_link.tickets
          if ticket.payload is not None)
    self.assertTupleEqual((test_front_to_back_datum,), front_to_back_payloads)
    self.assertTupleEqual((test_back_to_front_datum,), back_to_front_payloads)

  def _perform_scenario_test(self, scenario):
    test_operation_id = object()
    test_method = scenario.method()
    test_fore_link = _test_links.ForeLink(None, None)
    rear_lock = threading.Lock()
    rear_sequence_number = [0]
    def rear_action(front_to_back_ticket, fore_link):
      with rear_lock:
        if front_to_back_ticket.payload is not None:
          response = scenario.response_for_request(front_to_back_ticket.payload)
        else:
          response = None
      terminal = front_to_back_ticket.kind in (
          interfaces.FrontToBackTicket.Kind.COMPLETION,
          interfaces.FrontToBackTicket.Kind.ENTIRE)
      if response is not None or terminal:
        if terminal:
          kind = interfaces.BackToFrontTicket.Kind.COMPLETION
        else:
          kind = interfaces.BackToFrontTicket.Kind.CONTINUATION
        back_to_front_ticket = interfaces.BackToFrontTicket(
            front_to_back_ticket.operation_id, rear_sequence_number[0], kind,
            response)
        rear_sequence_number[0] += 1
        fore_link.accept_back_to_front_ticket(back_to_front_ticket)
    test_rear_link = _test_links.RearLink(rear_action, None)

    fore_link = fore.ForeLink(
        self.fore_link_pool, {test_method: scenario.deserialize_request},
        {test_method: scenario.serialize_response}, None, ())
    fore_link.join_rear_link(test_rear_link)
    test_rear_link.join_fore_link(fore_link)
    fore_link.start()
    port = fore_link.port()

    rear_link = rear.RearLink(
        'localhost', port, self.rear_link_pool,
        {test_method: scenario.serialize_request},
        {test_method: scenario.deserialize_response}, False, None, None, None)
    rear_link.join_fore_link(test_fore_link)
    test_fore_link.join_rear_link(rear_link)
    rear_link.start()

    commencement_ticket = interfaces.FrontToBackTicket(
        test_operation_id, 0,
        interfaces.FrontToBackTicket.Kind.COMMENCEMENT, test_method,
        interfaces.ServicedSubscription.Kind.FULL, None, None,
        _TIMEOUT)
    fore_sequence_number = 1
    rear_link.accept_front_to_back_ticket(commencement_ticket)
    for request in scenario.requests():
      continuation_ticket = interfaces.FrontToBackTicket(
          test_operation_id, fore_sequence_number,
          interfaces.FrontToBackTicket.Kind.CONTINUATION, None, None, None,
          request, None)
      fore_sequence_number += 1
      rear_link.accept_front_to_back_ticket(continuation_ticket)
    completion_ticket = interfaces.FrontToBackTicket(
        test_operation_id, fore_sequence_number,
        interfaces.FrontToBackTicket.Kind.COMPLETION, None, None, None, None,
        None)
    fore_sequence_number += 1
    rear_link.accept_front_to_back_ticket(completion_ticket)

    with test_fore_link.condition:
      while (not test_fore_link.tickets or
             test_fore_link.tickets[-1].kind is not
                 interfaces.BackToFrontTicket.Kind.COMPLETION):
        test_fore_link.condition.wait()

    rear_link.stop()
    fore_link.stop()

    with test_rear_link.condition:
      requests = tuple(
          ticket.payload for ticket in test_rear_link.tickets
          if ticket.payload is not None)
    with test_fore_link.condition:
      responses = tuple(
          ticket.payload for ticket in test_fore_link.tickets
          if ticket.payload is not None)
    self.assertTrue(scenario.verify_requests(requests))
    self.assertTrue(scenario.verify_responses(responses))

  def testEmptyScenario(self):
    self._perform_scenario_test(_proto_scenarios.EmptyScenario())

  def testBidirectionallyUnaryScenario(self):
    self._perform_scenario_test(_proto_scenarios.BidirectionallyUnaryScenario())

  def testBidirectionallyStreamingScenario(self):
    self._perform_scenario_test(
        _proto_scenarios.BidirectionallyStreamingScenario())


if __name__ == '__main__':
  unittest.main(verbosity=2)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""State and behavior for operation termination."""

import enum

from grpc.framework.base import _constants
from grpc.framework.base import _interfaces
from grpc.framework.base import interfaces
from grpc.framework.foundation import callable_util

_CALLBACK_EXCEPTION_LOG_MESSAGE = 'Exception calling termination callback!'


@enum.unique
class _Requirement(enum.Enum):
  """Symbols indicating events required for termination."""

  EMISSION = 'emission'
  TRANSMISSION = 'transmission'
  INGESTION = 'ingestion'

_FRONT_NOT_LISTENING_REQUIREMENTS = (_Requirement.TRANSMISSION,)
_BACK_NOT_LISTENING_REQUIREMENTS = (
    _Requirement.EMISSION, _Requirement.INGESTION,)
_LISTENING_REQUIREMENTS = (
    _Requirement.TRANSMISSION, _Requirement.INGESTION,)


class _TerminationManager(_interfaces.TerminationManager):
  """An implementation of _interfaces.TerminationManager."""

  def __init__(
      self, work_pool, utility_pool, action, requirements, local_failure):
    """Constructor.

    Args:
      work_pool: A thread pool in which customer work will be done.
      utility_pool: A thread pool in which work utility work will be done.
      action: An action to call on operation termination.
      requirements: A combination of _Requirement values identifying what
        must finish for the operation to be considered completed.
      local_failure: An interfaces.Outcome specifying what constitutes local
        failure of customer work.
    """
    self._work_pool = work_pool
    self._utility_pool = utility_pool
    self._action = action
    self._local_failure = local_failure
    self._has_locally_failed = False
    self._expiration_manager = None

    self._outstanding_requirements = set(requirements)
    self._outcome = None
    self._callbacks = []

  def set_expiration_manager(self, expiration_manager):
    self._expiration_manager = expiration_manager

  def _terminate(self, outcome):
    """Terminates the operation.

    Args:
      outcome: An interfaces.Outcome describing the outcome of the operation.
    """
    self._expiration_manager.abort()
    self._outstanding_requirements = None
    callbacks = list(self._callbacks)
    self._callbacks = None
    self._outcome = outcome

    act = callable_util.with_exceptions_logged(
        self._action, _constants.INTERNAL_ERROR_LOG_MESSAGE)

    if self._has_locally_failed:
      self._utility_pool.submit(act, outcome)
    else:
      def call_callbacks_and_act(callbacks, outcome):
        for callback in callbacks:
          callback_outcome = callable_util.call_logging_exceptions(
              callback, _CALLBACK_EXCEPTION_LOG_MESSAGE, outcome)
          if callback_outcome.exception is not None:
            outcome = self._local_failure
            break
        self._utility_pool.submit(act, outcome)

      self._work_pool.submit(callable_util.with_exceptions_logged(
          call_callbacks_and_act,
          _constants.INTERNAL_ERROR_LOG_MESSAGE),
                             callbacks, outcome)

  def is_active(self):
    """See _interfaces.TerminationManager.is_active for specification."""
    return self._outstanding_requirements is not None

  def add_callback(self, callback):
    """See _interfaces.TerminationManager.add_callback for specification."""
    if not self._has_locally_failed:
      if self._outstanding_requirements is None:
        self._work_pool.submit(
            callable_util.with_exceptions_logged(
                callback, _CALLBACK_EXCEPTION_LOG_MESSAGE), self._outcome)
      else:
        self._callbacks.append(callback)

  def emission_complete(self):
    """See superclass method for specification."""
    if self._outstanding_requirements is not None:
      self._outstanding_requirements.discard(_Requirement.EMISSION)
      if not self._outstanding_requirements:
        self._terminate(interfaces.Outcome.COMPLETED)

  def transmission_complete(self):
    """See superclass method for specification."""
    if self._outstanding_requirements is not None:
      self._outstanding_requirements.discard(_Requirement.TRANSMISSION)
      if not self._outstanding_requirements:
        self._terminate(interfaces.Outcome.COMPLETED)

  def ingestion_complete(self):
    """See superclass method for specification."""
    if self._outstanding_requirements is not None:
      self._outstanding_requirements.discard(_Requirement.INGESTION)
      if not self._outstanding_requirements:
        self._terminate(interfaces.Outcome.COMPLETED)

  def abort(self, outcome):
    """See _interfaces.TerminationManager.abort for specification."""
    if outcome is self._local_failure:
      self._has_failed_locally = True
    if self._outstanding_requirements is not None:
      self._terminate(outcome)


def front_termination_manager(
    work_pool, utility_pool, action, subscription_kind):
  """Creates a TerminationManager appropriate for front-side use.

  Args:
    work_pool: A thread pool in which customer work will be done.
    utility_pool: A thread pool in which work utility work will be done.
    action: An action to call on operation termination.
    subscription_kind: An interfaces.ServicedSubscription.Kind value.

  Returns:
    A TerminationManager appropriate for front-side use.
  """
  if subscription_kind is interfaces.ServicedSubscription.Kind.NONE:
    requirements = _FRONT_NOT_LISTENING_REQUIREMENTS
  else:
    requirements = _LISTENING_REQUIREMENTS

  return _TerminationManager(
      work_pool, utility_pool, action, requirements,
      interfaces.Outcome.SERVICED_FAILURE)


def back_termination_manager(work_pool, utility_pool, action, subscription_kind):
  """Creates a TerminationManager appropriate for back-side use.

  Args:
    work_pool: A thread pool in which customer work will be done.
    utility_pool: A thread pool in which work utility work will be done.
    action: An action to call on operation termination.
    subscription_kind: An interfaces.ServicedSubscription.Kind value.

  Returns:
    A TerminationManager appropriate for back-side use.
  """
  if subscription_kind is interfaces.ServicedSubscription.Kind.NONE:
    requirements = _BACK_NOT_LISTENING_REQUIREMENTS
  else:
    requirements = _LISTENING_REQUIREMENTS

  return _TerminationManager(
      work_pool, utility_pool, action, requirements,
      interfaces.Outcome.SERVICER_FAILURE)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utilities for working with callables."""

import abc
import collections
import enum
import functools
import logging


class Outcome(object):
  """A sum type describing the outcome of some call.

  Attributes:
    kind: One of Kind.RETURNED or Kind.RAISED respectively indicating that the
      call returned a value or raised an exception.
    return_value: The value returned by the call. Must be present if kind is
      Kind.RETURNED.
    exception: The exception raised by the call. Must be present if kind is
      Kind.RAISED.
  """
  __metaclass__ = abc.ABCMeta

  @enum.unique
  class Kind(enum.Enum):
    """Identifies the general kind of the outcome of some call."""

    RETURNED = object()
    RAISED = object()


class _EasyOutcome(
    collections.namedtuple(
        '_EasyOutcome', ['kind', 'return_value', 'exception']),
    Outcome):
  """A trivial implementation of Outcome."""


def _call_logging_exceptions(behavior, message, *args, **kwargs):
  try:
    return _EasyOutcome(Outcome.Kind.RETURNED, behavior(*args, **kwargs), None)
  except Exception as e:  # pylint: disable=broad-except
    logging.exception(message)
    return _EasyOutcome(Outcome.Kind.RAISED, None, e)


def with_exceptions_logged(behavior, message):
  """Wraps a callable in a try-except that logs any exceptions it raises.

  Args:
    behavior: Any callable.
    message: A string to log if the behavior raises an exception.

  Returns:
    A callable that when executed invokes the given behavior. The returned
      callable takes the same arguments as the given behavior but returns a
      future.Outcome describing whether the given behavior returned a value or
      raised an exception.
  """
  @functools.wraps(behavior)
  def wrapped_behavior(*args, **kwargs):
    return _call_logging_exceptions(behavior, message, *args, **kwargs)
  return wrapped_behavior


def call_logging_exceptions(behavior, message, *args, **kwargs):
  """Calls a behavior in a try-except that logs any exceptions it raises.

  Args:
    behavior: Any callable.
    message: A string to log if the behavior raises an exception.
    *args: Positional arguments to pass to the given behavior.
    **kwargs: Keyword arguments to pass to the given behavior.

  Returns:
    An Outcome describing whether the given behavior returned a value or raised
      an exception.
  """
  return _call_logging_exceptions(behavior, message, *args, **kwargs)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Enables scheduling execution at a later time."""

import time

from grpc.framework.foundation import _timer_future


def later(delay, computation):
  """Schedules later execution of a callable.

  Args:
    delay: Any numeric value. Represents the minimum length of time in seconds
      to allow to pass before beginning the computation. No guarantees are made
      about the maximum length of time that will pass.
    computation: A callable that accepts no arguments.

  Returns:
    A Future representing the scheduled computation.
  """
  timer_future = _timer_future.TimerFuture(time.time() + delay, computation)
  timer_future.start()
  return timer_future



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A test to verify an implementation of the Face layer of RPC Framework."""

import abc
import contextlib
import threading
import unittest

from grpc.framework.face import exceptions
from grpc.framework.face.testing import control
from grpc.framework.face.testing import coverage
from grpc.framework.face.testing import digest
from grpc.framework.face.testing import stock_service
from grpc.framework.face.testing import test_case
from grpc.framework.foundation import future
from grpc.framework.foundation import logging_pool

_TIMEOUT = 3
_MAXIMUM_POOL_SIZE = 10


class _PauseableIterator(object):

  def __init__(self, upstream):
    self._upstream = upstream
    self._condition = threading.Condition()
    self._paused = False

  @contextlib.contextmanager
  def pause(self):
    with self._condition:
      self._paused = True
    yield
    with self._condition:
      self._paused = False
      self._condition.notify_all()

  def __iter__(self):
    return self

  def next(self):
    with self._condition:
      while self._paused:
        self._condition.wait()
    return next(self._upstream)


class FutureInvocationAsynchronousEventServiceTestCase(
    test_case.FaceTestCase, coverage.FullCoverage):
  """A test of the Face layer of RPC Framework.

  Concrete subclasses must also extend unittest.TestCase.
  """
  __metaclass__ = abc.ABCMeta

  def setUp(self):
    """See unittest.TestCase.setUp for full specification.

    Overriding implementations must call this implementation.
    """
    self.control = control.PauseFailControl()
    self.digest_pool = logging_pool.pool(_MAXIMUM_POOL_SIZE)
    self.digest = digest.digest(
        stock_service.STOCK_TEST_SERVICE, self.control, self.digest_pool)

    self.stub, self.memo = self.set_up_implementation(
        self.digest.name, self.digest.methods,
        self.digest.event_method_implementations, None)

  def tearDown(self):
    """See unittest.TestCase.tearDown for full specification.

    Overriding implementations must call this implementation.
    """
    self.tear_down_implementation(self.memo)
    self.digest_pool.shutdown(wait=True)

  def testSuccessfulUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        response_future = self.stub.future_value_in_value_out(
            name, request, _TIMEOUT)
        response = response_future.result()

        test_messages.verify(request, response, self)

  def testSuccessfulUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        response_iterator = self.stub.inline_value_in_stream_out(
            name, request, _TIMEOUT)
        responses = list(response_iterator)

        test_messages.verify(request, responses, self)

  def testSuccessfulStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        request_iterator = _PauseableIterator(iter(requests))

        # Use of a paused iterator of requests allows us to test that control is
        # returned to calling code before the iterator yields any requests.
        with request_iterator.pause():
          response_future = self.stub.future_stream_in_value_out(
              name, request_iterator, _TIMEOUT)
        response = response_future.result()

        test_messages.verify(requests, response, self)

  def testSuccessfulStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        request_iterator = _PauseableIterator(iter(requests))

        # Use of a paused iterator of requests allows us to test that control is
        # returned to calling code before the iterator yields any requests.
        with request_iterator.pause():
          response_iterator = self.stub.inline_stream_in_stream_out(
              name, request_iterator, _TIMEOUT)
        responses = list(response_iterator)

        test_messages.verify(requests, responses, self)

  def testSequentialInvocations(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        first_request = test_messages.request()
        second_request = test_messages.request()

        first_response_future = self.stub.future_value_in_value_out(
            name, first_request, _TIMEOUT)
        first_response = first_response_future.result()

        test_messages.verify(first_request, first_response, self)

        second_response_future = self.stub.future_value_in_value_out(
            name, second_request, _TIMEOUT)
        second_response = second_response_future.result()

        test_messages.verify(second_request, second_response, self)

  def testExpiredUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        with self.control.pause():
          multi_callable = self.stub.unary_unary_multi_callable(name)
          response_future = multi_callable.future(request, _TIMEOUT)
          self.assertIsInstance(
              response_future.exception(), exceptions.ExpirationError)
          with self.assertRaises(exceptions.ExpirationError):
            response_future.result()

  def testExpiredUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        with self.control.pause():
          response_iterator = self.stub.inline_value_in_stream_out(
              name, request, _TIMEOUT)
          with self.assertRaises(exceptions.ExpirationError):
            list(response_iterator)

  def testExpiredStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        with self.control.pause():
          multi_callable = self.stub.stream_unary_multi_callable(name)
          response_future = multi_callable.future(iter(requests), _TIMEOUT)
          self.assertIsInstance(
              response_future.exception(), exceptions.ExpirationError)
          with self.assertRaises(exceptions.ExpirationError):
            response_future.result()

  def testExpiredStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        with self.control.pause():
          response_iterator = self.stub.inline_stream_in_stream_out(
              name, iter(requests), _TIMEOUT)
          with self.assertRaises(exceptions.ExpirationError):
            list(response_iterator)

  def testFailedUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        with self.control.fail():
          response_future = self.stub.future_value_in_value_out(
              name, request, _TIMEOUT)

          # Because the servicer fails outside of the thread from which the
          # servicer-side runtime called into it its failure is
          # indistinguishable from simply not having called its
          # response_callback before the expiration of the RPC.
          self.assertIsInstance(
              response_future.exception(), exceptions.ExpirationError)
          with self.assertRaises(exceptions.ExpirationError):
            response_future.result()

  def testFailedUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        # Because the servicer fails outside of the thread from which the
        # servicer-side runtime called into it its failure is indistinguishable
        # from simply not having called its response_consumer before the
        # expiration of the RPC.
        with self.control.fail(), self.assertRaises(exceptions.ExpirationError):
          response_iterator = self.stub.inline_value_in_stream_out(
              name, request, _TIMEOUT)
          list(response_iterator)

  def testFailedStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        with self.control.fail():
          response_future = self.stub.future_stream_in_value_out(
              name, iter(requests), _TIMEOUT)

          # Because the servicer fails outside of the thread from which the
          # servicer-side runtime called into it its failure is
          # indistinguishable from simply not having called its
          # response_callback before the expiration of the RPC.
          self.assertIsInstance(
              response_future.exception(), exceptions.ExpirationError)
          with self.assertRaises(exceptions.ExpirationError):
            response_future.result()

  def testFailedStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        # Because the servicer fails outside of the thread from which the
        # servicer-side runtime called into it its failure is indistinguishable
        # from simply not having called its response_consumer before the
        # expiration of the RPC.
        with self.control.fail(), self.assertRaises(exceptions.ExpirationError):
          response_iterator = self.stub.inline_stream_in_stream_out(
              name, iter(requests), _TIMEOUT)
          list(response_iterator)

  def testParallelInvocations(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        first_request = test_messages.request()
        second_request = test_messages.request()

        first_response_future = self.stub.future_value_in_value_out(
            name, first_request, _TIMEOUT)
        second_response_future = self.stub.future_value_in_value_out(
            name, second_request, _TIMEOUT)
        first_response = first_response_future.result()
        second_response = second_response_future.result()

        test_messages.verify(first_request, first_response, self)
        test_messages.verify(second_request, second_response, self)

  @unittest.skip('TODO(nathaniel): implement.')
  def testWaitingForSomeButNotAllParallelInvocations(self):
    raise NotImplementedError()

  def testCancelledUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        with self.control.pause():
          response_future = self.stub.future_value_in_value_out(
              name, request, _TIMEOUT)
          cancel_method_return_value = response_future.cancel()

        self.assertFalse(cancel_method_return_value)
        self.assertTrue(response_future.cancelled())

  def testCancelledUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()

        with self.control.pause():
          response_iterator = self.stub.inline_value_in_stream_out(
              name, request, _TIMEOUT)
          response_iterator.cancel()

        with self.assertRaises(future.CancelledError):
          next(response_iterator)

  def testCancelledStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        with self.control.pause():
          response_future = self.stub.future_stream_in_value_out(
              name, iter(requests), _TIMEOUT)
          cancel_method_return_value = response_future.cancel()

        self.assertFalse(cancel_method_return_value)
        self.assertTrue(response_future.cancelled())

  def testCancelledStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()

        with self.control.pause():
          response_iterator = self.stub.inline_stream_in_stream_out(
              name, iter(requests), _TIMEOUT)
          response_iterator.cancel()

        with self.assertRaises(future.CancelledError):
          next(response_iterator)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
import collections
import enum

# TODO(atash): decide whether or not to move these enums to the _c module to
# force build errors with upstream changes.

class GrpcChannelArgumentKeys(enum.Enum):
  """Mirrors keys used in grpc_channel_args for GRPC-specific arguments."""
  SSL_TARGET_NAME_OVERRIDE = 'grpc.ssl_target_name_override'

@enum.unique
class CallError(enum.IntEnum):
  """Mirrors grpc_call_error in the C core."""
  OK                        = 0
  ERROR                     = 1
  ERROR_NOT_ON_SERVER       = 2
  ERROR_NOT_ON_CLIENT       = 3
  ERROR_ALREADY_ACCEPTED    = 4
  ERROR_ALREADY_INVOKED     = 5
  ERROR_NOT_INVOKED         = 6
  ERROR_ALREADY_FINISHED    = 7
  ERROR_TOO_MANY_OPERATIONS = 8
  ERROR_INVALID_FLAGS       = 9
  ERROR_INVALID_METADATA    = 10

@enum.unique
class StatusCode(enum.IntEnum):
  """Mirrors grpc_status_code in the C core."""
  OK                  = 0
  CANCELLED           = 1
  UNKNOWN             = 2
  INVALID_ARGUMENT    = 3
  DEADLINE_EXCEEDED   = 4
  NOT_FOUND           = 5
  ALREADY_EXISTS      = 6
  PERMISSION_DENIED   = 7
  RESOURCE_EXHAUSTED  = 8
  FAILED_PRECONDITION = 9
  ABORTED             = 10
  OUT_OF_RANGE        = 11
  UNIMPLEMENTED       = 12
  INTERNAL            = 13
  UNAVAILABLE         = 14
  DATA_LOSS           = 15
  UNAUTHENTICATED     = 16

@enum.unique
class OpType(enum.IntEnum):
  """Mirrors grpc_op_type in the C core."""
  SEND_INITIAL_METADATA   = 0
  SEND_MESSAGE            = 1
  SEND_CLOSE_FROM_CLIENT  = 2
  SEND_STATUS_FROM_SERVER = 3
  RECV_INITIAL_METADATA   = 4
  RECV_MESSAGE            = 5
  RECV_STATUS_ON_CLIENT   = 6
  RECV_CLOSE_ON_SERVER    = 7

@enum.unique
class EventType(enum.IntEnum):
  """Mirrors grpc_completion_type in the C core."""
  QUEUE_SHUTDOWN  = 0
  QUEUE_TIMEOUT   = 1  # if seen on the Python side, something went horridly wrong
  OP_COMPLETE     = 2

class Status(collections.namedtuple(
    'Status', [
        'code',
        'details',
    ])):
  """The end status of a GRPC call.

  Attributes:
    code (StatusCode): ...
    details (str): ...
  """

class CallDetails(collections.namedtuple(
    'CallDetails', [
        'method',
        'host',
        'deadline',
    ])):
  """Provides information to the server about the client's call.

  Attributes:
    method (str): ...
    host (str): ...
    deadline (float): ...
  """

class OpArgs(collections.namedtuple(
    'OpArgs', [
        'type',
        'initial_metadata',
        'trailing_metadata',
        'message',
        'status',
    ])):
  """Arguments passed into a GRPC operation.

  Attributes:
    type (OpType): ...
    initial_metadata (sequence of 2-sequence of str): Only valid if type ==
      OpType.SEND_INITIAL_METADATA, else is None.
    trailing_metadata (sequence of 2-sequence of str): Only valid if type ==
      OpType.SEND_STATUS_FROM_SERVER, else is None.
    message (bytes): Only valid if type == OpType.SEND_MESSAGE, else is None.
    status (Status): Only valid if type == OpType.SEND_STATUS_FROM_SERVER, else
      is None.
  """

  @staticmethod
  def send_initial_metadata(initial_metadata):
    return OpArgs(OpType.SEND_INITIAL_METADATA, initial_metadata, None, None, None)

  @staticmethod
  def send_message(message):
    return OpArgs(OpType.SEND_MESSAGE, None, None, message, None)

  @staticmethod
  def send_close_from_client():
    return OpArgs(OpType.SEND_CLOSE_FROM_CLIENT, None, None, None, None)

  @staticmethod
  def send_status_from_server(trailing_metadata, status_code, status_details):
    return OpArgs(OpType.SEND_STATUS_FROM_SERVER, None, trailing_metadata, None, Status(status_code, status_details))

  @staticmethod
  def recv_initial_metadata():
    return OpArgs(OpType.RECV_INITIAL_METADATA, None, None, None, None);

  @staticmethod
  def recv_message():
    return OpArgs(OpType.RECV_MESSAGE, None, None, None, None)

  @staticmethod
  def recv_status_on_client():
    return OpArgs(OpType.RECV_STATUS_ON_CLIENT, None, None, None, None)

  @staticmethod
  def recv_close_on_server():
    return OpArgs(OpType.RECV_CLOSE_ON_SERVER, None, None, None, None)


class OpResult(collections.namedtuple(
    'OpResult', [
        'type',
        'initial_metadata',
        'trailing_metadata',
        'message',
        'status',
        'cancelled',
    ])):
  """Results received from a GRPC operation.

  Attributes:
    type (OpType): ...
    initial_metadata (sequence of 2-sequence of str): Only valid if type ==
      OpType.RECV_INITIAL_METADATA, else is None.
    trailing_metadata (sequence of 2-sequence of str): Only valid if type ==
      OpType.RECV_STATUS_ON_CLIENT, else is None.
    message (bytes): Only valid if type == OpType.RECV_MESSAGE, else is None.
    status (Status): Only valid if type == OpType.RECV_STATUS_ON_CLIENT, else
      is None.
    cancelled (bool): Only valid if type == OpType.RECV_CLOSE_ON_SERVER, else
      is None.
  """


class Event(collections.namedtuple(
    'Event', [
        'type',
        'tag',
        'call',
        'call_details',
        'results',
        'success',
    ])):
  """An event received from a GRPC completion queue.

  Attributes:
    type (EventType): ...
    tag (object): ...
    call (Call): The Call object associated with this event (if there is one,
      else None).
    call_details (CallDetails): The call details associated with the
      server-side call (if there is such information, else None).
    results (list of OpResult): ...
    success (bool): ...
  """


class CompletionQueue:
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __init__(self):
    pass

  def __iter__(self):
    """This class may be iterated over.

    This is the equivalent of calling next() repeatedly with an absolute
    deadline of None (i.e. no deadline).
    """
    return self

  @abc.abstractmethod
  def next(self, deadline=float('+inf')):
    """Get the next event on this completion queue.

    Args:
      deadline (float): absolute deadline in seconds from the Python epoch, or
        None for no deadline.

    Returns:
      Event: ...
    """
    pass

  @abc.abstractmethod
  def shutdown(self):
    """Begin the shutdown process of this completion queue.

    Note that this does not immediately destroy the completion queue.
    Nevertheless, user code should not pass it around after invoking this.
    """
    return None


class Call:
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def start_batch(self, ops, tag):
    """Start a batch of operations.

    Args:
      ops (sequence of OpArgs): ...
      tag (object): ...

    Returns:
      CallError: ...
    """
    return CallError.ERROR

  @abc.abstractmethod
  def cancel(self, code=None, details=None):
    """Cancel the call.

    Args:
      code (int): Status code to cancel with (on the server side). If
        specified, so must `details`.
      details (str): Status details to cancel with (on the server side). If
        specified, so must `code`.

    Returns:
      CallError: ...
    """
    return CallError.ERROR


class Channel:
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __init__(self, target, args, credentials=None):
    """Initialize a Channel.

    Args:
      target (str): ...
      args (sequence of 2-sequence of str, (str|integer)): ...
      credentials (ClientCredentials): If None, create an insecure channel,
        else create a secure channel using the client credentials.
    """

  @abc.abstractmethod
  def create_call(self, completion_queue, method, host, deadline=float('+inf')):
    """Create a call from this channel.

    Args:
      completion_queue (CompletionQueue): ...
      method (str): ...
      host (str): ...
      deadline (float): absolute deadline in seconds from the Python epoch, or
        None for no deadline.

    Returns:
      Call: call object associated with this Channel and passed parameters.
    """
    return None


class Server:
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __init__(self, completion_queue, args):
    """Initialize a server.

    Args:
      completion_queue (CompletionQueue): ...
      args (sequence of 2-sequence of str, (str|integer)): ...
    """

  @abc.abstractmethod
  def add_http2_port(self, address, credentials=None):
    """Adds an HTTP/2 address+port to the server.

    Args:
      address (str): ...
      credentials (ServerCredentials): If None, create an insecure port, else
        create a secure port using the server credentials.
    """

  @abc.abstractmethod
  def start(self):
    """Starts the server."""

  @abc.abstractmethod
  def shutdown(self, tag=None):
    """Shuts down the server. Does not immediately destroy the server.

    Args:
      tag (object): if not None, have the server place an event on its
        completion queue notifying it when this server has completely shut down.
    """

  @abc.abstractmethod
  def request_call(self, completion_queue, tag):
    """Requests a call from the server on the server's completion queue.

    Args:
      completion_queue (CompletionQueue): Completion queue for the call. May be
        the same as the server's completion queue.
      tag (object) ...
    """



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Links suitable for use in tests."""

import threading

from grpc.framework.base import interfaces


class ForeLink(interfaces.ForeLink):
  """A ForeLink suitable for use in tests of RearLinks."""

  def __init__(self, action, rear_link):
    self.condition = threading.Condition()
    self.tickets = []
    self.action = action
    self.rear_link = rear_link

  def accept_back_to_front_ticket(self, ticket):
    with self.condition:
      self.tickets.append(ticket)
      self.condition.notify_all()
      action, rear_link = self.action, self.rear_link

    if action is not None:
      action(ticket, rear_link)

  def join_rear_link(self, rear_link):
    with self.condition:
      self.rear_link = rear_link


class RearLink(interfaces.RearLink):
  """A RearLink suitable for use in tests of ForeLinks."""

  def __init__(self, action, fore_link):
    self.condition = threading.Condition()
    self.tickets = []
    self.action = action
    self.fore_link = fore_link

  def accept_front_to_back_ticket(self, ticket):
    with self.condition:
      self.tickets.append(ticket)
      self.condition.notify_all()
      action, fore_link = self.action, self.fore_link

    if action is not None:
      action(ticket, fore_link)

  def join_fore_link(self, fore_link):
    with self.condition:
      self.fore_link = fore_link



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from grpc._adapter import _c
from grpc._adapter import _types

ClientCredentials = _c.ClientCredentials
ServerCredentials = _c.ServerCredentials


class CompletionQueue(_types.CompletionQueue):

  def __init__(self):
    self.completion_queue = _c.CompletionQueue()

  def next(self, deadline=float('+inf')):
    raw_event = self.completion_queue.next(deadline)
    if raw_event is None:
      return None
    event = _types.Event(*raw_event)
    if event.call is not None:
      event = event._replace(call=Call(event.call))
    if event.call_details is not None:
      event = event._replace(call_details=_types.CallDetails(*event.call_details))
    if event.results is not None:
      new_results = [_types.OpResult(*r) for r in event.results]
      new_results = [r if r.status is None else r._replace(status=_types.Status(_types.StatusCode(r.status[0]), r.status[1])) for r in new_results]
      event = event._replace(results=new_results)
    return event

  def shutdown(self):
    self.completion_queue.shutdown()


class Call(_types.Call):

  def __init__(self, call):
    self.call = call

  def start_batch(self, ops, tag):
    return self.call.start_batch(ops, tag)

  def cancel(self, code=None, details=None):
    if code is None and details is None:
      return self.call.cancel()
    else:
      return self.call.cancel(code, details)


class Channel(_types.Channel):

  def __init__(self, target, args, creds=None):
    if creds is None:
      self.channel = _c.Channel(target, args)
    else:
      self.channel = _c.Channel(target, args, creds)

  def create_call(self, completion_queue, method, host, deadline=None):
    return Call(self.channel.create_call(completion_queue.completion_queue, method, host, deadline))


_NO_TAG = object()

class Server(_types.Server):

  def __init__(self, completion_queue, args):
    self.server = _c.Server(completion_queue.completion_queue, args)

  def add_http2_port(self, addr, creds=None):
    if creds is None:
      return self.server.add_http2_port(addr)
    else:
      return self.server.add_http2_port(addr, creds)

  def start(self):
    return self.server.start()

  def shutdown(self, tag=None):
    return self.server.shutdown(tag)

  def request_call(self, completion_queue, tag):
    return self.server.request_call(completion_queue.completion_queue, tag)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Implementations of FrontLinks and BackLinks."""

import collections
import threading
import uuid

# _interfaces is referenced from specification in this module.
from grpc.framework.base import _cancellation
from grpc.framework.base import _context
from grpc.framework.base import _emission
from grpc.framework.base import _expiration
from grpc.framework.base import _ingestion
from grpc.framework.base import _interfaces  # pylint: disable=unused-import
from grpc.framework.base import _reception
from grpc.framework.base import _termination
from grpc.framework.base import _transmission
from grpc.framework.base import interfaces
from grpc.framework.foundation import callable_util

_IDLE_ACTION_EXCEPTION_LOG_MESSAGE = 'Exception calling idle action!'


class _EasyOperation(interfaces.Operation):
  """A trivial implementation of interfaces.Operation."""

  def __init__(self, emission_manager, context, cancellation_manager):
    """Constructor.

    Args:
      emission_manager: The _interfaces.EmissionManager for the operation that
        will accept values emitted by customer code.
      context: The interfaces.OperationContext for use by the customer
        during the operation.
      cancellation_manager: The _interfaces.CancellationManager for the
        operation.
    """
    self.consumer = emission_manager
    self.context = context
    self._cancellation_manager = cancellation_manager

  def cancel(self):
    self._cancellation_manager.cancel()


class _Endlette(object):
  """Utility for stateful behavior common to Fronts and Backs."""

  def __init__(self, pool):
    """Constructor.

    Args:
      pool: A thread pool to use when calling registered idle actions.
    """
    self._lock = threading.Lock()
    self._pool = pool
    # Dictionary from operation IDs to ReceptionManager-or-None. A None value
    # indicates an in-progress fire-and-forget operation for which the customer
    # has chosen to ignore results.
    self._operations = {}
    self._stats = {outcome: 0 for outcome in interfaces.Outcome}
    self._idle_actions = []

  def terminal_action(self, operation_id):
    """Constructs the termination action for a single operation.

    Args:
      operation_id: An operation ID.

    Returns:
      A callable that takes an operation outcome for an argument to be used as
        the termination action for the operation associated with the given
        operation ID.
    """
    def termination_action(outcome):
      with self._lock:
        self._stats[outcome] += 1
        self._operations.pop(operation_id, None)
        if not self._operations:
          for action in self._idle_actions:
            self._pool.submit(callable_util.with_exceptions_logged(
                action, _IDLE_ACTION_EXCEPTION_LOG_MESSAGE))
          self._idle_actions = []
    return termination_action

  def __enter__(self):
    self._lock.acquire()

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._lock.release()

  def get_operation(self, operation_id):
    return self._operations.get(operation_id, None)

  def add_operation(self, operation_id, operation_reception_manager):
    self._operations[operation_id] = operation_reception_manager

  def operation_stats(self):
    with self._lock:
      return dict(self._stats)

  def add_idle_action(self, action):
    with self._lock:
      if self._operations:
        self._idle_actions.append(action)
      else:
        self._pool.submit(callable_util.with_exceptions_logged(
            action, _IDLE_ACTION_EXCEPTION_LOG_MESSAGE))


class _FrontManagement(
    collections.namedtuple(
        '_FrontManagement',
        ('reception', 'emission', 'operation', 'cancellation'))):
  """Just a trivial helper class to bundle four fellow-traveling objects."""


def _front_operate(
    callback, work_pool, transmission_pool, utility_pool,
    termination_action, operation_id, name, payload, complete, timeout,
    subscription, trace_id):
  """Constructs objects necessary for front-side operation management.

  Args:
    callback: A callable that accepts interfaces.FrontToBackTickets and
      delivers them to the other side of the operation. Execution of this
      callable may take any arbitrary length of time.
    work_pool: A thread pool in which to execute customer code.
    transmission_pool: A thread pool to use for transmitting to the other side
      of the operation.
    utility_pool: A thread pool for utility tasks.
    termination_action: A no-arg behavior to be called upon operation
      completion.
    operation_id: An object identifying the operation.
    name: The name of the method being called during the operation.
    payload: The first customer-significant value to be transmitted to the other
      side. May be None if there is no such value or if the customer chose not
      to pass it at operation invocation.
    complete: A boolean indicating whether or not additional payloads will be
      supplied by the customer.
    timeout: A length of time in seconds to allow for the operation.
    subscription: A interfaces.ServicedSubscription describing the
      customer's interest in the results of the operation.
    trace_id: A uuid.UUID identifying a set of related operations to which this
      operation belongs. May be None.

  Returns:
    A _FrontManagement object bundling together the
      _interfaces.ReceptionManager, _interfaces.EmissionManager,
      _context.OperationContext, and _interfaces.CancellationManager for the
      operation.
  """
  lock = threading.Lock()
  with lock:
    termination_manager = _termination.front_termination_manager(
        work_pool, utility_pool, termination_action, subscription.kind)
    transmission_manager = _transmission.front_transmission_manager(
        lock, transmission_pool, callback, operation_id, name,
        subscription.kind, trace_id, timeout, termination_manager)
    operation_context = _context.OperationContext(
        lock, operation_id, interfaces.Outcome.SERVICED_FAILURE,
        termination_manager, transmission_manager)
    emission_manager = _emission.front_emission_manager(
        lock, termination_manager, transmission_manager)
    ingestion_manager = _ingestion.front_ingestion_manager(
        lock, work_pool, subscription, termination_manager,
        transmission_manager, operation_context)
    expiration_manager = _expiration.front_expiration_manager(
        lock, termination_manager, transmission_manager, ingestion_manager,
        timeout)
    reception_manager = _reception.front_reception_manager(
        lock, termination_manager, transmission_manager, ingestion_manager,
        expiration_manager)
    cancellation_manager = _cancellation.CancellationManager(
        lock, termination_manager, transmission_manager, ingestion_manager,
        expiration_manager)

    termination_manager.set_expiration_manager(expiration_manager)
    transmission_manager.set_ingestion_and_expiration_managers(
        ingestion_manager, expiration_manager)
    operation_context.set_ingestion_and_expiration_managers(
        ingestion_manager, expiration_manager)
    emission_manager.set_ingestion_manager_and_expiration_manager(
        ingestion_manager, expiration_manager)
    ingestion_manager.set_expiration_manager(expiration_manager)

    transmission_manager.inmit(payload, complete)

    if subscription.kind is interfaces.ServicedSubscription.Kind.NONE:
      returned_reception_manager = None
    else:
      returned_reception_manager = reception_manager

    return _FrontManagement(
        returned_reception_manager, emission_manager, operation_context,
        cancellation_manager)


class FrontLink(interfaces.FrontLink):
  """An implementation of interfaces.FrontLink."""

  def __init__(self, work_pool, transmission_pool, utility_pool):
    """Constructor.

    Args:
      work_pool: A thread pool to be used for executing customer code.
      transmission_pool: A thread pool to be used for transmitting values to
        the other side of the operation.
      utility_pool: A thread pool to be used for utility tasks.
    """
    self._endlette = _Endlette(utility_pool)
    self._work_pool = work_pool
    self._transmission_pool = transmission_pool
    self._utility_pool = utility_pool
    self._callback = None

    self._operations = {}

  def join_rear_link(self, rear_link):
    """See interfaces.ForeLink.join_rear_link for specification."""
    with self._endlette:
      self._callback = rear_link.accept_front_to_back_ticket

  def operation_stats(self):
    """See interfaces.End.operation_stats for specification."""
    return self._endlette.operation_stats()

  def add_idle_action(self, action):
    """See interfaces.End.add_idle_action for specification."""
    self._endlette.add_idle_action(action)

  def operate(
      self, name, payload, complete, timeout, subscription, trace_id):
    """See interfaces.Front.operate for specification."""
    operation_id = uuid.uuid4()
    with self._endlette:
      management = _front_operate(
          self._callback, self._work_pool, self._transmission_pool,
          self._utility_pool, self._endlette.terminal_action(operation_id),
          operation_id, name, payload, complete, timeout, subscription,
          trace_id)
      self._endlette.add_operation(operation_id, management.reception)
      return _EasyOperation(
          management.emission, management.operation, management.cancellation)

  def accept_back_to_front_ticket(self, ticket):
    """See interfaces.End.act for specification."""
    with self._endlette:
      reception_manager = self._endlette.get_operation(ticket.operation_id)
    if reception_manager:
      reception_manager.receive_ticket(ticket)


def _back_operate(
    servicer, callback, work_pool, transmission_pool, utility_pool,
    termination_action, ticket, default_timeout, maximum_timeout):
  """Constructs objects necessary for back-side operation management.

  Also begins back-side operation by feeding the first received ticket into the
  constructed _interfaces.ReceptionManager.

  Args:
    servicer: An interfaces.Servicer for servicing operations.
    callback: A callable that accepts interfaces.BackToFrontTickets and
      delivers them to the other side of the operation. Execution of this
      callable may take any arbitrary length of time.
    work_pool: A thread pool in which to execute customer code.
    transmission_pool: A thread pool to use for transmitting to the other side
      of the operation.
    utility_pool: A thread pool for utility tasks.
    termination_action: A no-arg behavior to be called upon operation
      completion.
    ticket: The first interfaces.FrontToBackTicket received for the operation.
    default_timeout: A length of time in seconds to be used as the default
      time alloted for a single operation.
    maximum_timeout: A length of time in seconds to be used as the maximum
      time alloted for a single operation.

  Returns:
    The _interfaces.ReceptionManager to be used for the operation.
  """
  lock = threading.Lock()
  with lock:
    termination_manager = _termination.back_termination_manager(
        work_pool, utility_pool, termination_action, ticket.subscription)
    transmission_manager = _transmission.back_transmission_manager(
        lock, transmission_pool, callback, ticket.operation_id,
        termination_manager, ticket.subscription)
    operation_context = _context.OperationContext(
        lock, ticket.operation_id, interfaces.Outcome.SERVICER_FAILURE,
        termination_manager, transmission_manager)
    emission_manager = _emission.back_emission_manager(
        lock, termination_manager, transmission_manager)
    ingestion_manager = _ingestion.back_ingestion_manager(
        lock, work_pool, servicer, termination_manager,
        transmission_manager, operation_context, emission_manager)
    expiration_manager = _expiration.back_expiration_manager(
        lock, termination_manager, transmission_manager, ingestion_manager,
        ticket.timeout, default_timeout, maximum_timeout)
    reception_manager = _reception.back_reception_manager(
        lock, termination_manager, transmission_manager, ingestion_manager,
        expiration_manager)

    termination_manager.set_expiration_manager(expiration_manager)
    transmission_manager.set_ingestion_and_expiration_managers(
        ingestion_manager, expiration_manager)
    operation_context.set_ingestion_and_expiration_managers(
        ingestion_manager, expiration_manager)
    emission_manager.set_ingestion_manager_and_expiration_manager(
        ingestion_manager, expiration_manager)
    ingestion_manager.set_expiration_manager(expiration_manager)

  reception_manager.receive_ticket(ticket)

  return reception_manager


class BackLink(interfaces.BackLink):
  """An implementation of interfaces.BackLink."""

  def __init__(
      self, servicer, work_pool, transmission_pool, utility_pool,
      default_timeout, maximum_timeout):
    """Constructor.

    Args:
      servicer: An interfaces.Servicer for servicing operations.
      work_pool: A thread pool in which to execute customer code.
      transmission_pool: A thread pool to use for transmitting to the other side
        of the operation.
      utility_pool: A thread pool for utility tasks.
      default_timeout: A length of time in seconds to be used as the default
        time alloted for a single operation.
      maximum_timeout: A length of time in seconds to be used as the maximum
        time alloted for a single operation.
    """
    self._endlette = _Endlette(utility_pool)
    self._servicer = servicer
    self._work_pool = work_pool
    self._transmission_pool = transmission_pool
    self._utility_pool = utility_pool
    self._default_timeout = default_timeout
    self._maximum_timeout = maximum_timeout
    self._callback = None

  def join_fore_link(self, fore_link):
    """See interfaces.RearLink.join_fore_link for specification."""
    with self._endlette:
      self._callback = fore_link.accept_back_to_front_ticket

  def accept_front_to_back_ticket(self, ticket):
    """See interfaces.RearLink.accept_front_to_back_ticket for specification."""
    with self._endlette:
      reception_manager = self._endlette.get_operation(ticket.operation_id)
      if reception_manager is None:
        reception_manager = _back_operate(
            self._servicer, self._callback, self._work_pool,
            self._transmission_pool, self._utility_pool,
            self._endlette.terminal_action(ticket.operation_id), ticket,
            self._default_timeout, self._maximum_timeout)
        self._endlette.add_operation(ticket.operation_id, reception_manager)
      else:
        reception_manager.receive_ticket(ticket)

  def operation_stats(self):
    """See interfaces.End.operation_stats for specification."""
    return self._endlette.operation_stats()

  def add_idle_action(self, action):
    """See interfaces.End.add_idle_action for specification."""
    self._endlette.add_idle_action(action)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utility functions for invoking RPCs."""

import sys
import threading

from grpc.framework.base import interfaces as base_interfaces
from grpc.framework.base import util as base_util
from grpc.framework.face import _control
from grpc.framework.face import interfaces
from grpc.framework.foundation import callable_util
from grpc.framework.foundation import future

_ITERATOR_EXCEPTION_LOG_MESSAGE = 'Exception iterating over requests!'
_DONE_CALLBACK_LOG_MESSAGE = 'Exception calling Future "done" callback!'


class _RendezvousServicedIngestor(base_interfaces.ServicedIngestor):

  def __init__(self, rendezvous):
    self._rendezvous = rendezvous

  def consumer(self, operation_context):
    return self._rendezvous


class _EventServicedIngestor(base_interfaces.ServicedIngestor):

  def __init__(self, result_consumer, abortion_callback):
    self._result_consumer = result_consumer
    self._abortion_callback = abortion_callback

  def consumer(self, operation_context):
    operation_context.add_termination_callback(
        _control.as_operation_termination_callback(self._abortion_callback))
    return self._result_consumer


def _rendezvous_subscription(rendezvous):
  return base_util.full_serviced_subscription(
      _RendezvousServicedIngestor(rendezvous))


def _unary_event_subscription(completion_callback, abortion_callback):
  return base_util.full_serviced_subscription(
      _EventServicedIngestor(
          _control.UnaryConsumer(completion_callback), abortion_callback))


def _stream_event_subscription(result_consumer, abortion_callback):
  return base_util.full_serviced_subscription(
      _EventServicedIngestor(result_consumer, abortion_callback))


# NOTE(nathaniel): This class has some extremely special semantics around
# cancellation that allow it to be used by both "blocking" APIs and "futures"
# APIs.
#
# Since futures.Future defines its own exception for cancellation, we want these
# objects, when returned by methods of a returning-Futures-from-other-methods
# object, to raise the same exception for cancellation. But that's weird in a
# blocking API - why should this object, also returned by methods of blocking
# APIs, raise exceptions from the "future" module? Should we do something like
# have this class be parameterized by the type of exception that it raises in
# cancellation circumstances?
#
# We don't have to take such a dramatic step: since blocking APIs define no
# cancellation semantics whatsoever, there is no supported way for
# blocking-API-users of these objects to cancel RPCs, and thus no supported way
# for them to see an exception the type of which would be weird to them.
#
# Bonus: in both blocking and futures APIs, this object still properly raises
# exceptions.CancellationError for any *server-side cancellation* of an RPC.
class _OperationCancellableIterator(interfaces.CancellableIterator):
  """An interfaces.CancellableIterator for response-streaming operations."""

  def __init__(self, rendezvous, operation):
    self._lock = threading.Lock()
    self._rendezvous = rendezvous
    self._operation = operation
    self._cancelled = False

  def __iter__(self):
    return self

  def next(self):
    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
    return next(self._rendezvous)

  def cancel(self):
    with self._lock:
      self._cancelled = True
    self._operation.cancel()
    self._rendezvous.set_outcome(base_interfaces.Outcome.CANCELLED)


class _OperationFuture(future.Future):
  """A future.Future interface to an operation."""

  def __init__(self, rendezvous, operation):
    self._condition = threading.Condition()
    self._rendezvous = rendezvous
    self._operation = operation

    self._cancelled = False
    self._computed = False
    self._payload = None
    self._exception = None
    self._traceback = None
    self._callbacks = []

  def cancel(self):
    """See future.Future.cancel for specification."""
    with self._condition:
      if not self._cancelled and not self._computed:
        self._operation.cancel()
        self._cancelled = True
        self._condition.notify_all()
    return False

  def cancelled(self):
    """See future.Future.cancelled for specification."""
    with self._condition:
      return self._cancelled

  def running(self):
    """See future.Future.running for specification."""
    with self._condition:
      return not self._cancelled and not self._computed

  def done(self):
    """See future.Future.done for specification."""
    with self._condition:
      return self._cancelled or self._computed

  def result(self, timeout=None):
    """See future.Future.result for specification."""
    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      if self._computed:
        if self._payload is None:
          raise self._exception  # pylint: disable=raising-bad-type
        else:
          return self._payload

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._callbacks.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        if self._payload is None:
          raise self._exception  # pylint: disable=raising-bad-type
        else:
          return self._payload
      else:
        raise future.TimeoutError()

  def exception(self, timeout=None):
    """See future.Future.exception for specification."""
    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      if self._computed:
        return self._exception

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._callbacks.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._exception
      else:
        raise future.TimeoutError()

  def traceback(self, timeout=None):
    """See future.Future.traceback for specification."""
    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      if self._computed:
        return self._traceback

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._callbacks.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._condition:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._traceback
      else:
        raise future.TimeoutError()

  def add_done_callback(self, fn):
    """See future.Future.add_done_callback for specification."""
    with self._condition:
      if self._callbacks is not None:
        self._callbacks.append(fn)
        return

    callable_util.call_logging_exceptions(fn, _DONE_CALLBACK_LOG_MESSAGE, self)

  def on_operation_termination(self, operation_outcome):
    """Indicates to this object that the operation has terminated.

    Args:
      operation_outcome: A base_interfaces.Outcome value indicating the
        outcome of the operation.
    """
    with self._condition:
      cancelled = self._cancelled
      if cancelled:
        callbacks = list(self._callbacks)
        self._callbacks = None
      else:
        rendezvous = self._rendezvous

    if not cancelled:
      payload = None
      exception = None
      traceback = None
      if operation_outcome == base_interfaces.Outcome.COMPLETED:
        try:
          payload = next(rendezvous)
        except Exception as e:  # pylint: disable=broad-except
          exception = e
          traceback = sys.exc_info()[2]
      else:
        try:
          # We raise and then immediately catch in order to create a traceback.
          raise _control.abortion_outcome_to_exception(operation_outcome)
        except Exception as e:  # pylint: disable=broad-except
          exception = e
          traceback = sys.exc_info()[2]
      with self._condition:
        if not self._cancelled:
          self._computed = True
          self._payload = payload
          self._exception = exception
          self._traceback = traceback
        callbacks = list(self._callbacks)
        self._callbacks = None

    for callback in callbacks:
      callable_util.call_logging_exceptions(
          callback, _DONE_CALLBACK_LOG_MESSAGE, self)


class _Call(interfaces.Call):

  def __init__(self, operation):
    self._operation = operation
    self.context = _control.RpcContext(operation.context)

  def cancel(self):
    self._operation.cancel()


def blocking_value_in_value_out(front, name, payload, timeout, trace_id):
  """Services in a blocking fashion a value-in value-out servicer method."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(
      name, payload, True, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  return next(rendezvous)


def future_value_in_value_out(front, name, payload, timeout, trace_id):
  """Services a value-in value-out servicer method by returning a Future."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(
      name, payload, True, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  operation_future = _OperationFuture(rendezvous, operation)
  operation.context.add_termination_callback(
      operation_future.on_operation_termination)
  return operation_future


def inline_value_in_stream_out(front, name, payload, timeout, trace_id):
  """Services a value-in stream-out servicer method."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(
      name, payload, True, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  return _OperationCancellableIterator(rendezvous, operation)


def blocking_stream_in_value_out(
    front, name, payload_iterator, timeout, trace_id):
  """Services in a blocking fashion a stream-in value-out servicer method."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(name, None, False, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  for payload in payload_iterator:
    operation.consumer.consume(payload)
  operation.consumer.terminate()
  return next(rendezvous)


def future_stream_in_value_out(
    front, name, payload_iterator, timeout, trace_id, pool):
  """Services a stream-in value-out servicer method by returning a Future."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(name, None, False, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  pool.submit(
      callable_util.with_exceptions_logged(
          _control.pipe_iterator_to_consumer, _ITERATOR_EXCEPTION_LOG_MESSAGE),
      payload_iterator, operation.consumer, lambda: True, True)
  operation_future = _OperationFuture(rendezvous, operation)
  operation.context.add_termination_callback(
      operation_future.on_operation_termination)
  return operation_future


def inline_stream_in_stream_out(
    front, name, payload_iterator, timeout, trace_id, pool):
  """Services a stream-in stream-out servicer method."""
  rendezvous = _control.Rendezvous()
  subscription = _rendezvous_subscription(rendezvous)
  operation = front.operate(name, None, False, timeout, subscription, trace_id)
  operation.context.add_termination_callback(rendezvous.set_outcome)
  pool.submit(
      callable_util.with_exceptions_logged(
          _control.pipe_iterator_to_consumer, _ITERATOR_EXCEPTION_LOG_MESSAGE),
      payload_iterator, operation.consumer, lambda: True, True)
  return _OperationCancellableIterator(rendezvous, operation)


def event_value_in_value_out(
    front, name, payload, completion_callback, abortion_callback, timeout,
    trace_id):
  subscription = _unary_event_subscription(
      completion_callback, abortion_callback)
  operation = front.operate(
      name, payload, True, timeout, subscription, trace_id)
  return _Call(operation)


def event_value_in_stream_out(
    front, name, payload, result_payload_consumer, abortion_callback, timeout,
    trace_id):
  subscription = _stream_event_subscription(
      result_payload_consumer, abortion_callback)
  operation = front.operate(
      name, payload, True, timeout, subscription, trace_id)
  return _Call(operation)


def event_stream_in_value_out(
    front, name, completion_callback, abortion_callback, timeout, trace_id):
  subscription = _unary_event_subscription(
      completion_callback, abortion_callback)
  operation = front.operate(name, None, False, timeout, subscription, trace_id)
  return _Call(operation), operation.consumer


def event_stream_in_stream_out(
    front, name, result_payload_consumer, abortion_callback, timeout, trace_id):
  subscription = _stream_event_subscription(
      result_payload_consumer, abortion_callback)
  operation = front.operate(name, None, False, timeout, subscription, trace_id)
  return _Call(operation), operation.consumer



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utilities for RPC framework's face layer."""

import collections

from grpc.framework.common import cardinality
from grpc.framework.common import style
from grpc.framework.face import interfaces
from grpc.framework.foundation import stream


class _MethodImplementation(
    interfaces.MethodImplementation,
    collections.namedtuple(
        '_MethodImplementation',
        ['cardinality', 'style', 'unary_unary_inline', 'unary_stream_inline',
         'stream_unary_inline', 'stream_stream_inline', 'unary_unary_event',
         'unary_stream_event', 'stream_unary_event', 'stream_stream_event',])):
  pass


def unary_unary_inline(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a unary-unary RPC method as a callable value
      that takes a request value and an interfaces.RpcContext object and
      returns a response value.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.UNARY_UNARY, style.Service.INLINE, behavior,
      None, None, None, None, None, None, None)


def unary_stream_inline(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a unary-stream RPC method as a callable
      value that takes a request value and an interfaces.RpcContext object and
      returns an iterator of response values.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.UNARY_STREAM, style.Service.INLINE, None,
      behavior, None, None, None, None, None, None)


def stream_unary_inline(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a stream-unary RPC method as a callable
      value that takes an iterator of request values and an
      interfaces.RpcContext object and returns a response value.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.STREAM_UNARY, style.Service.INLINE, None, None,
      behavior, None, None, None, None, None)


def stream_stream_inline(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a stream-stream RPC method as a callable
      value that takes an iterator of request values and an
      interfaces.RpcContext object and returns an iterator of response values.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.STREAM_STREAM, style.Service.INLINE, None, None,
      None, behavior, None, None, None, None)


def unary_unary_event(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a unary-unary RPC method as a callable
      value that takes a request value, a response callback to which to pass
      the response value of the RPC, and an interfaces.RpcContext.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.UNARY_UNARY, style.Service.EVENT, None, None,
      None, None, behavior, None, None, None)


def unary_stream_event(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a unary-stream RPC method as a callable
      value that takes a request value, a stream.Consumer to which to pass the
      the response values of the RPC, and an interfaces.RpcContext.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.UNARY_STREAM, style.Service.EVENT, None, None,
      None, None, None, behavior, None, None)


def stream_unary_event(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a stream-unary RPC method as a callable
      value that takes a response callback to which to pass the response value
      of the RPC and an interfaces.RpcContext and returns a stream.Consumer to
      which the request values of the RPC should be passed.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.STREAM_UNARY, style.Service.EVENT, None, None,
      None, None, None, None, behavior, None)


def stream_stream_event(behavior):
  """Creates an interfaces.MethodImplementation for the given behavior.

  Args:
    behavior: The implementation of a stream-stream RPC method as a callable
      value that takes a stream.Consumer to which to pass the response values
      of the RPC and an interfaces.RpcContext and returns a stream.Consumer to
      which the request values of the RPC should be passed.

  Returns:
    An interfaces.MethodImplementation derived from the given behavior.
  """
  return _MethodImplementation(
      cardinality.Cardinality.STREAM_STREAM, style.Service.EVENT, None, None,
      None, None, None, None, None, behavior)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for the old '_low'."""

import Queue
import threading
import time
import unittest

from grpc._adapter import _intermediary_low as _low

_STREAM_LENGTH = 300
_TIMEOUT = 5
_AFTER_DELAY = 2
_FUTURE = time.time() + 60 * 60 * 24
_BYTE_SEQUENCE = b'\abcdefghijklmnopqrstuvwxyz0123456789' * 200
_BYTE_SEQUENCE_SEQUENCE = tuple(
    bytes(bytearray((row + column) % 256 for column in range(row)))
    for row in range(_STREAM_LENGTH))


class LonelyClientTest(unittest.TestCase):

  def testLonelyClient(self):
    host = 'nosuchhostexists'
    port = 54321
    method = 'test method'
    deadline = time.time() + _TIMEOUT
    after_deadline = deadline + _AFTER_DELAY
    metadata_tag = object()
    finish_tag = object()

    completion_queue = _low.CompletionQueue()
    channel = _low.Channel('%s:%d' % (host, port), None)
    client_call = _low.Call(channel, completion_queue, method, host, deadline)

    client_call.invoke(completion_queue, metadata_tag, finish_tag)
    first_event = completion_queue.get(after_deadline)
    self.assertIsNotNone(first_event)
    second_event = completion_queue.get(after_deadline)
    self.assertIsNotNone(second_event)
    kinds = [event.kind for event in (first_event, second_event)]
    self.assertItemsEqual(
        (_low.Event.Kind.METADATA_ACCEPTED, _low.Event.Kind.FINISH),
        kinds)

    self.assertIsNone(completion_queue.get(after_deadline))

    completion_queue.stop()
    stop_event = completion_queue.get(_FUTURE)
    self.assertEqual(_low.Event.Kind.STOP, stop_event.kind)

    del client_call
    del channel
    del completion_queue


def _drive_completion_queue(completion_queue, event_queue):
  while True:
    event = completion_queue.get(_FUTURE)
    if event.kind is _low.Event.Kind.STOP:
      break
    event_queue.put(event)


class EchoTest(unittest.TestCase):

  def setUp(self):
    self.host = 'localhost'

    self.server_completion_queue = _low.CompletionQueue()
    self.server = _low.Server(self.server_completion_queue)
    port = self.server.add_http2_addr('[::]:0')
    self.server.start()
    self.server_events = Queue.Queue()
    self.server_completion_queue_thread = threading.Thread(
        target=_drive_completion_queue,
        args=(self.server_completion_queue, self.server_events))
    self.server_completion_queue_thread.start()

    self.client_completion_queue = _low.CompletionQueue()
    self.channel = _low.Channel('%s:%d' % (self.host, port), None)
    self.client_events = Queue.Queue()
    self.client_completion_queue_thread = threading.Thread(
        target=_drive_completion_queue,
        args=(self.client_completion_queue, self.client_events))
    self.client_completion_queue_thread.start()

  def tearDown(self):
    self.server.stop()
    self.server_completion_queue.stop()
    self.client_completion_queue.stop()
    self.server_completion_queue_thread.join()
    self.client_completion_queue_thread.join()
    del self.server

  def _perform_echo_test(self, test_data):
    method = 'test method'
    details = 'test details'
    server_leading_metadata_key = 'my_server_leading_key'
    server_leading_metadata_value = 'my_server_leading_value'
    server_trailing_metadata_key = 'my_server_trailing_key'
    server_trailing_metadata_value = 'my_server_trailing_value'
    client_metadata_key = 'my_client_key'
    client_metadata_value = 'my_client_value'
    server_leading_binary_metadata_key = 'my_server_leading_key-bin'
    server_leading_binary_metadata_value = b'\0'*2047
    server_trailing_binary_metadata_key = 'my_server_trailing_key-bin'
    server_trailing_binary_metadata_value = b'\0'*2047
    client_binary_metadata_key = 'my_client_key-bin'
    client_binary_metadata_value = b'\0'*2047
    deadline = _FUTURE
    metadata_tag = object()
    finish_tag = object()
    write_tag = object()
    complete_tag = object()
    service_tag = object()
    read_tag = object()
    status_tag = object()

    server_data = []
    client_data = []

    client_call = _low.Call(self.channel, self.client_completion_queue,
                            method, self.host, deadline)
    client_call.add_metadata(client_metadata_key, client_metadata_value)
    client_call.add_metadata(client_binary_metadata_key,
                             client_binary_metadata_value)

    client_call.invoke(self.client_completion_queue, metadata_tag, finish_tag)

    self.server.service(service_tag)
    service_accepted = self.server_events.get()
    self.assertIsNotNone(service_accepted)
    self.assertIs(service_accepted.kind, _low.Event.Kind.SERVICE_ACCEPTED)
    self.assertIs(service_accepted.tag, service_tag)
    self.assertEqual(method, service_accepted.service_acceptance.method)
    self.assertEqual(self.host, service_accepted.service_acceptance.host)
    self.assertIsNotNone(service_accepted.service_acceptance.call)
    metadata = dict(service_accepted.metadata)
    self.assertIn(client_metadata_key, metadata)
    self.assertEqual(client_metadata_value, metadata[client_metadata_key])
    self.assertIn(client_binary_metadata_key, metadata)
    self.assertEqual(client_binary_metadata_value,
                     metadata[client_binary_metadata_key])
    server_call = service_accepted.service_acceptance.call
    server_call.accept(self.server_completion_queue, finish_tag)
    server_call.add_metadata(server_leading_metadata_key,
                             server_leading_metadata_value)
    server_call.add_metadata(server_leading_binary_metadata_key,
                             server_leading_binary_metadata_value)
    server_call.premetadata()

    metadata_accepted = self.client_events.get()
    self.assertIsNotNone(metadata_accepted)
    self.assertEqual(_low.Event.Kind.METADATA_ACCEPTED, metadata_accepted.kind)
    self.assertEqual(metadata_tag, metadata_accepted.tag)
    metadata = dict(metadata_accepted.metadata)
    self.assertIn(server_leading_metadata_key, metadata)
    self.assertEqual(server_leading_metadata_value,
                     metadata[server_leading_metadata_key])
    self.assertIn(server_leading_binary_metadata_key, metadata)
    self.assertEqual(server_leading_binary_metadata_value,
                     metadata[server_leading_binary_metadata_key])

    for datum in test_data:
      client_call.write(datum, write_tag)
      write_accepted = self.client_events.get()
      self.assertIsNotNone(write_accepted)
      self.assertIs(write_accepted.kind, _low.Event.Kind.WRITE_ACCEPTED)
      self.assertIs(write_accepted.tag, write_tag)
      self.assertIs(write_accepted.write_accepted, True)

      server_call.read(read_tag)
      read_accepted = self.server_events.get()
      self.assertIsNotNone(read_accepted)
      self.assertEqual(_low.Event.Kind.READ_ACCEPTED, read_accepted.kind)
      self.assertEqual(read_tag, read_accepted.tag)
      self.assertIsNotNone(read_accepted.bytes)
      server_data.append(read_accepted.bytes)

      server_call.write(read_accepted.bytes, write_tag)
      write_accepted = self.server_events.get()
      self.assertIsNotNone(write_accepted)
      self.assertEqual(_low.Event.Kind.WRITE_ACCEPTED, write_accepted.kind)
      self.assertEqual(write_tag, write_accepted.tag)
      self.assertTrue(write_accepted.write_accepted)

      client_call.read(read_tag)
      read_accepted = self.client_events.get()
      self.assertIsNotNone(read_accepted)
      self.assertEqual(_low.Event.Kind.READ_ACCEPTED, read_accepted.kind)
      self.assertEqual(read_tag, read_accepted.tag)
      self.assertIsNotNone(read_accepted.bytes)
      client_data.append(read_accepted.bytes)

    client_call.complete(complete_tag)
    complete_accepted = self.client_events.get()
    self.assertIsNotNone(complete_accepted)
    self.assertIs(complete_accepted.kind, _low.Event.Kind.COMPLETE_ACCEPTED)
    self.assertIs(complete_accepted.tag, complete_tag)
    self.assertIs(complete_accepted.complete_accepted, True)

    server_call.read(read_tag)
    read_accepted = self.server_events.get()
    self.assertIsNotNone(read_accepted)
    self.assertEqual(_low.Event.Kind.READ_ACCEPTED, read_accepted.kind)
    self.assertEqual(read_tag, read_accepted.tag)
    self.assertIsNone(read_accepted.bytes)

    server_call.add_metadata(server_trailing_metadata_key,
                             server_trailing_metadata_value)
    server_call.add_metadata(server_trailing_binary_metadata_key,
                             server_trailing_binary_metadata_value)

    server_call.status(_low.Status(_low.Code.OK, details), status_tag)
    server_terminal_event_one = self.server_events.get()
    server_terminal_event_two = self.server_events.get()
    if server_terminal_event_one.kind == _low.Event.Kind.COMPLETE_ACCEPTED:
      status_accepted = server_terminal_event_one
      rpc_accepted = server_terminal_event_two
    else:
      status_accepted = server_terminal_event_two
      rpc_accepted = server_terminal_event_one
    self.assertIsNotNone(status_accepted)
    self.assertIsNotNone(rpc_accepted)
    self.assertEqual(_low.Event.Kind.COMPLETE_ACCEPTED, status_accepted.kind)
    self.assertEqual(status_tag, status_accepted.tag)
    self.assertTrue(status_accepted.complete_accepted)
    self.assertEqual(_low.Event.Kind.FINISH, rpc_accepted.kind)
    self.assertEqual(finish_tag, rpc_accepted.tag)
    self.assertEqual(_low.Status(_low.Code.OK, ''), rpc_accepted.status)

    client_call.read(read_tag)
    client_terminal_event_one = self.client_events.get()
    client_terminal_event_two = self.client_events.get()
    if client_terminal_event_one.kind == _low.Event.Kind.READ_ACCEPTED:
      read_accepted = client_terminal_event_one
      finish_accepted = client_terminal_event_two
    else:
      read_accepted = client_terminal_event_two
      finish_accepted = client_terminal_event_one
    self.assertIsNotNone(read_accepted)
    self.assertIsNotNone(finish_accepted)
    self.assertEqual(_low.Event.Kind.READ_ACCEPTED, read_accepted.kind)
    self.assertEqual(read_tag, read_accepted.tag)
    self.assertIsNone(read_accepted.bytes)
    self.assertEqual(_low.Event.Kind.FINISH, finish_accepted.kind)
    self.assertEqual(finish_tag, finish_accepted.tag)
    self.assertEqual(_low.Status(_low.Code.OK, details), finish_accepted.status)
    metadata = dict(finish_accepted.metadata)
    self.assertIn(server_trailing_metadata_key, metadata)
    self.assertEqual(server_trailing_metadata_value,
                     metadata[server_trailing_metadata_key])
    self.assertIn(server_trailing_binary_metadata_key, metadata)
    self.assertEqual(server_trailing_binary_metadata_value,
                     metadata[server_trailing_binary_metadata_key])
    self.assertSetEqual(set(key for key, _ in finish_accepted.metadata),
                        set((server_trailing_metadata_key,
                             server_trailing_binary_metadata_key,)))

    server_timeout_none_event = self.server_completion_queue.get(0)
    self.assertIsNone(server_timeout_none_event)
    client_timeout_none_event = self.client_completion_queue.get(0)
    self.assertIsNone(client_timeout_none_event)

    self.assertSequenceEqual(test_data, server_data)
    self.assertSequenceEqual(test_data, client_data)

  def testNoEcho(self):
    self._perform_echo_test(())

  def testOneByteEcho(self):
    self._perform_echo_test([b'\x07'])

  def testOneManyByteEcho(self):
    self._perform_echo_test([_BYTE_SEQUENCE])

  def testManyOneByteEchoes(self):
    self._perform_echo_test(_BYTE_SEQUENCE)

  def testManyManyByteEchoes(self):
    self._perform_echo_test(_BYTE_SEQUENCE_SEQUENCE)


class CancellationTest(unittest.TestCase):

  def setUp(self):
    self.host = 'localhost'

    self.server_completion_queue = _low.CompletionQueue()
    self.server = _low.Server(self.server_completion_queue)
    port = self.server.add_http2_addr('[::]:0')
    self.server.start()
    self.server_events = Queue.Queue()
    self.server_completion_queue_thread = threading.Thread(
        target=_drive_completion_queue,
        args=(self.server_completion_queue, self.server_events))
    self.server_completion_queue_thread.start()

    self.client_completion_queue = _low.CompletionQueue()
    self.channel = _low.Channel('%s:%d' % (self.host, port), None)
    self.client_events = Queue.Queue()
    self.client_completion_queue_thread = threading.Thread(
        target=_drive_completion_queue,
        args=(self.client_completion_queue, self.client_events))
    self.client_completion_queue_thread.start()

  def tearDown(self):
    self.server.stop()
    self.server_completion_queue.stop()
    self.client_completion_queue.stop()
    self.server_completion_queue_thread.join()
    self.client_completion_queue_thread.join()
    del self.server

  def testCancellation(self):
    method = 'test method'
    deadline = _FUTURE
    metadata_tag = object()
    finish_tag = object()
    write_tag = object()
    service_tag = object()
    read_tag = object()
    test_data = _BYTE_SEQUENCE_SEQUENCE

    server_data = []
    client_data = []

    client_call = _low.Call(self.channel, self.client_completion_queue,
                            method, self.host, deadline)

    client_call.invoke(self.client_completion_queue, metadata_tag, finish_tag)

    self.server.service(service_tag)
    service_accepted = self.server_events.get()
    server_call = service_accepted.service_acceptance.call

    server_call.accept(self.server_completion_queue, finish_tag)
    server_call.premetadata()

    metadata_accepted = self.client_events.get()
    self.assertIsNotNone(metadata_accepted)

    for datum in test_data:
      client_call.write(datum, write_tag)
      write_accepted = self.client_events.get()

      server_call.read(read_tag)
      read_accepted = self.server_events.get()
      server_data.append(read_accepted.bytes)

      server_call.write(read_accepted.bytes, write_tag)
      write_accepted = self.server_events.get()
      self.assertIsNotNone(write_accepted)

      client_call.read(read_tag)
      read_accepted = self.client_events.get()
      client_data.append(read_accepted.bytes)

    client_call.cancel()
    # cancel() is idempotent.
    client_call.cancel()
    client_call.cancel()
    client_call.cancel()

    server_call.read(read_tag)

    server_terminal_event_one = self.server_events.get()
    server_terminal_event_two = self.server_events.get()
    if server_terminal_event_one.kind == _low.Event.Kind.READ_ACCEPTED:
      read_accepted = server_terminal_event_one
      rpc_accepted = server_terminal_event_two
    else:
      read_accepted = server_terminal_event_two
      rpc_accepted = server_terminal_event_one
    self.assertIsNotNone(read_accepted)
    self.assertIsNotNone(rpc_accepted)
    self.assertEqual(_low.Event.Kind.READ_ACCEPTED, read_accepted.kind)
    self.assertIsNone(read_accepted.bytes)
    self.assertEqual(_low.Event.Kind.FINISH, rpc_accepted.kind)
    self.assertEqual(_low.Status(_low.Code.CANCELLED, ''), rpc_accepted.status)

    finish_event = self.client_events.get()
    self.assertEqual(_low.Event.Kind.FINISH, finish_event.kind)
    self.assertEqual(_low.Status(_low.Code.CANCELLED, 'Cancelled'), 
                                 finish_event.status)

    server_timeout_none_event = self.server_completion_queue.get(0)
    self.assertIsNone(server_timeout_none_event)
    client_timeout_none_event = self.client_completion_queue.get(0)
    self.assertIsNone(client_timeout_none_event)

    self.assertSequenceEqual(test_data, server_data)
    self.assertSequenceEqual(test_data, client_data)


class ExpirationTest(unittest.TestCase):

  @unittest.skip('TODO(nathaniel): Expiration test!')
  def testExpiration(self):
    pass


if __name__ == '__main__':
  unittest.main(verbosity=2)




# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
import collections

# face_interfaces is referenced from specification in this module.
from grpc.framework.common import cardinality
from grpc.framework.face import interfaces as face_interfaces  # pylint: disable=unused-import
from grpc.framework.face import utilities as face_utilities
from grpc.framework.alpha import _reexport
from grpc.framework.alpha import interfaces


def _qualified_name(service_name, method_name):
  return '/%s/%s' % (service_name, method_name)


# TODO(nathaniel): This structure is getting bloated; it could be shrunk if
# implementations._Stub used a generic rather than a dynamic underlying
# face-layer stub.
class InvocationBreakdown(object):
  """An intermediate representation of invocation-side views of RPC methods.

  Attributes:
    cardinalities: A dictionary from RPC method name to interfaces.Cardinality
      value.
    qualified_names: A dictionary from unqualified RPC method name to
      service-qualified RPC method name.
    face_cardinalities: A dictionary from service-qualified RPC method name to
      to cardinality.Cardinality value.
    request_serializers: A dictionary from service-qualified RPC method name to
      callable behavior to be used serializing request values for the RPC.
    response_deserializers: A dictionary from service-qualified RPC method name
      to callable behavior to be used deserializing response values for the
      RPC.
  """
  __metaclass__ = abc.ABCMeta


class _EasyInvocationBreakdown(
    InvocationBreakdown,
    collections.namedtuple(
        '_EasyInvocationBreakdown',
        ('cardinalities', 'qualified_names', 'face_cardinalities',
         'request_serializers', 'response_deserializers'))):
  pass


class ServiceBreakdown(object):
  """An intermediate representation of service-side views of RPC methods.

  Attributes:
    implementations: A dictionary from service-qualified RPC method name to
      face_interfaces.MethodImplementation implementing the RPC method.
    request_deserializers: A dictionary from service-qualified RPC method name
      to callable behavior to be used deserializing request values for the RPC.
    response_serializers: A dictionary from service-qualified RPC method name
      to callable behavior to be used serializing response values for the RPC.
  """
  __metaclass__ = abc.ABCMeta


class _EasyServiceBreakdown(
    ServiceBreakdown,
    collections.namedtuple(
        '_EasyServiceBreakdown',
        ('implementations', 'request_deserializers', 'response_serializers'))):
  pass


def break_down_invocation(service_name, method_descriptions):
  """Derives an InvocationBreakdown from several RPC method descriptions.

  Args:
    service_name: The package-qualified full name of the service.
    method_descriptions: A dictionary from RPC method name to
      interfaces.RpcMethodInvocationDescription describing the RPCs.

  Returns:
    An InvocationBreakdown corresponding to the given method descriptions.
  """
  cardinalities = {}
  qualified_names = {}
  face_cardinalities = {}
  request_serializers = {}
  response_deserializers = {}
  for name, method_description in method_descriptions.iteritems():
    qualified_name = _qualified_name(service_name, name)
    method_cardinality = method_description.cardinality()
    cardinalities[name] = method_description.cardinality()
    qualified_names[name] = qualified_name
    face_cardinalities[qualified_name] = _reexport.common_cardinality(
        method_cardinality)
    request_serializers[qualified_name] = method_description.serialize_request
    response_deserializers[qualified_name] = (
        method_description.deserialize_response)
  return _EasyInvocationBreakdown(
      cardinalities, qualified_names, face_cardinalities, request_serializers,
      response_deserializers)


def break_down_service(service_name, method_descriptions):
  """Derives a ServiceBreakdown from several RPC method descriptions.

  Args:
    method_descriptions: A dictionary from RPC method name to
      interfaces.RpcMethodServiceDescription describing the RPCs.

  Returns:
    A ServiceBreakdown corresponding to the given method descriptions.
  """
  implementations = {}
  request_deserializers = {}
  response_serializers = {}
  for name, method_description in method_descriptions.iteritems():
    qualified_name = _qualified_name(service_name, name)
    method_cardinality = method_description.cardinality()
    if method_cardinality is interfaces.Cardinality.UNARY_UNARY:
      def service(
          request, face_rpc_context,
          service_behavior=method_description.service_unary_unary):
        return service_behavior(
            request, _reexport.rpc_context(face_rpc_context))
      implementations[qualified_name] = face_utilities.unary_unary_inline(
          service)
    elif method_cardinality is interfaces.Cardinality.UNARY_STREAM:
      def service(
          request, face_rpc_context,
          service_behavior=method_description.service_unary_stream):
        return service_behavior(
            request, _reexport.rpc_context(face_rpc_context))
      implementations[qualified_name] = face_utilities.unary_stream_inline(
          service)
    elif method_cardinality is interfaces.Cardinality.STREAM_UNARY:
      def service(
          request_iterator, face_rpc_context,
          service_behavior=method_description.service_stream_unary):
        return service_behavior(
            request_iterator, _reexport.rpc_context(face_rpc_context))
      implementations[qualified_name] = face_utilities.stream_unary_inline(
          service)
    elif method_cardinality is interfaces.Cardinality.STREAM_STREAM:
      def service(
          request_iterator, face_rpc_context,
          service_behavior=method_description.service_stream_stream):
        return service_behavior(
            request_iterator, _reexport.rpc_context(face_rpc_context))
      implementations[qualified_name] = face_utilities.stream_stream_inline(
          service)
    request_deserializers[qualified_name] = (
        method_description.deserialize_request)
    response_serializers[qualified_name] = (
        method_description.serialize_response)

  return _EasyServiceBreakdown(
      implementations, request_deserializers, response_serializers)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Interfaces defined and used by the base layer of RPC Framework."""

import abc
import collections
import enum

# stream is referenced from specification in this module.
from grpc.framework.foundation import stream  # pylint: disable=unused-import


@enum.unique
class Outcome(enum.Enum):
  """Operation outcomes."""

  COMPLETED = 'completed'
  CANCELLED = 'cancelled'
  EXPIRED = 'expired'
  RECEPTION_FAILURE = 'reception failure'
  TRANSMISSION_FAILURE = 'transmission failure'
  SERVICER_FAILURE = 'servicer failure'
  SERVICED_FAILURE = 'serviced failure'


class OperationContext(object):
  """Provides operation-related information and action.

  Attributes:
    trace_id: A uuid.UUID identifying a particular set of related operations.
  """
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def is_active(self):
    """Describes whether the operation is active or has terminated."""
    raise NotImplementedError()

  @abc.abstractmethod
  def add_termination_callback(self, callback):
    """Adds a function to be called upon operation termination.

    Args:
      callback: A callable that will be passed an Outcome value.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def time_remaining(self):
    """Describes the length of allowed time remaining for the operation.

    Returns:
      A nonnegative float indicating the length of allowed time in seconds
      remaining for the operation to complete before it is considered to have
      timed out.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def fail(self, exception):
    """Indicates that the operation has failed.

    Args:
      exception: An exception germane to the operation failure. May be None.
    """
    raise NotImplementedError()


class Servicer(object):
  """Interface for service implementations."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def service(self, name, context, output_consumer):
    """Services an operation.

    Args:
      name: The name of the operation.
      context: A ServicerContext object affording contextual information and
        actions.
      output_consumer: A stream.Consumer that will accept output values of
        the operation.

    Returns:
      A stream.Consumer that will accept input values for the operation.

    Raises:
      exceptions.NoSuchMethodError: If this Servicer affords no method with the
        given name.
      abandonment.Abandoned: If the operation has been aborted and there no
        longer is any reason to service the operation.
    """
    raise NotImplementedError()


class Operation(object):
  """Representation of an in-progress operation.

  Attributes:
    consumer: A stream.Consumer into which payloads constituting the operation's
      input may be passed.
    context: An OperationContext affording information and action about the
      operation.
  """
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def cancel(self):
    """Cancels this operation."""
    raise NotImplementedError()


class ServicedIngestor(object):
  """Responsible for accepting the result of an operation."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def consumer(self, operation_context):
    """Affords a consumer to which operation results will be passed.

    Args:
      operation_context: An OperationContext object for the current operation.

    Returns:
      A stream.Consumer to which the results of the current operation will be
        passed.

    Raises:
      abandonment.Abandoned: If the operation has been aborted and there no
        longer is any reason to service the operation.
    """
    raise NotImplementedError()


class ServicedSubscription(object):
  """A sum type representing a serviced's interest in an operation.

  Attributes:
    kind: A Kind value.
    ingestor: A ServicedIngestor. Must be present if kind is Kind.FULL. Must
      be None if kind is Kind.TERMINATION_ONLY or Kind.NONE.
  """
  __metaclass__ = abc.ABCMeta

  @enum.unique
  class Kind(enum.Enum):
    """Kinds of subscription."""

    FULL = 'full'
    TERMINATION_ONLY = 'termination only'
    NONE = 'none'


class End(object):
  """Common type for entry-point objects on both sides of an operation."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def operation_stats(self):
    """Reports the number of terminated operations broken down by outcome.

    Returns:
      A dictionary from Outcome value to an integer identifying the number
        of operations that terminated with that outcome.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def add_idle_action(self, action):
    """Adds an action to be called when this End has no ongoing operations.

    Args:
      action: A callable that accepts no arguments.
    """
    raise NotImplementedError()


class Front(End):
  """Clientish objects that afford the invocation of operations."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def operate(
      self, name, payload, complete, timeout, subscription, trace_id):
    """Commences an operation.

    Args:
      name: The name of the method invoked for the operation.
      payload: An initial payload for the operation. May be None.
      complete: A boolean indicating whether or not additional payloads to be
        sent to the servicer may be supplied after this call.
      timeout: A length of time in seconds to allow for the operation.
      subscription: A ServicedSubscription for the operation.
      trace_id: A uuid.UUID identifying a set of related operations to which
        this operation belongs.

    Returns:
      An Operation object affording information and action about the operation
        in progress.
    """
    raise NotImplementedError()


class Back(End):
  """Serverish objects that perform the work of operations."""
  __metaclass__ = abc.ABCMeta


class FrontToBackTicket(
    collections.namedtuple(
        'FrontToBackTicket',
        ['operation_id', 'sequence_number', 'kind', 'name', 'subscription',
         'trace_id', 'payload', 'timeout'])):
  """A sum type for all values sent from a front to a back.

  Attributes:
    operation_id: A unique-with-respect-to-equality hashable object identifying
      a particular operation.
    sequence_number: A zero-indexed integer sequence number identifying the
      ticket's place among all the tickets sent from front to back for this
      particular operation. Must be zero if kind is Kind.COMMENCEMENT or
      Kind.ENTIRE. Must be positive for any other kind.
    kind: A Kind value describing the overall kind of ticket.
    name: The name of an operation. Must be present if kind is Kind.COMMENCEMENT
      or Kind.ENTIRE. Must be None for any other kind.
    subscription: An ServicedSubscription.Kind value describing the interest
      the front has in tickets sent from the back. Must be present if
      kind is Kind.COMMENCEMENT or Kind.ENTIRE. Must be None for any other kind.
    trace_id: A uuid.UUID identifying a set of related operations to which this
      operation belongs. May be None.
    payload: A customer payload object. Must be present if kind is
      Kind.CONTINUATION. Must be None if kind is Kind.CANCELLATION. May be None
      for any other kind.
    timeout: An optional length of time (measured from the beginning of the
      operation) to allow for the entire operation. If None, a default value on
      the back will be used. If present and excessively large, the back may
      limit the operation to a smaller duration of its choice. May be present
      for any ticket kind; setting a value on a later ticket allows fronts
      to request time extensions (or even time reductions!) on in-progress
      operations.
  """

  @enum.unique
  class Kind(enum.Enum):
    """Identifies the overall kind of a FrontToBackTicket."""

    COMMENCEMENT = 'commencement'
    CONTINUATION = 'continuation'
    COMPLETION = 'completion'
    ENTIRE = 'entire'
    CANCELLATION = 'cancellation'
    EXPIRATION = 'expiration'
    SERVICER_FAILURE = 'servicer failure'
    SERVICED_FAILURE = 'serviced failure'
    RECEPTION_FAILURE = 'reception failure'
    TRANSMISSION_FAILURE = 'transmission failure'


class BackToFrontTicket(
    collections.namedtuple(
        'BackToFrontTicket',
        ['operation_id', 'sequence_number', 'kind', 'payload'])):
  """A sum type for all values sent from a back to a front.

  Attributes:
    operation_id: A unique-with-respect-to-equality hashable object identifying
      a particular operation.
    sequence_number: A zero-indexed integer sequence number identifying the
      ticket's place among all the tickets sent from back to front for this
      particular operation.
    kind: A Kind value describing the overall kind of ticket.
    payload: A customer payload object. Must be present if kind is
      Kind.CONTINUATION. May be None if kind is Kind.COMPLETION. Must be None
      otherwise.
  """

  @enum.unique
  class Kind(enum.Enum):
    """Identifies the overall kind of a BackToFrontTicket."""

    CONTINUATION = 'continuation'
    COMPLETION = 'completion'
    CANCELLATION = 'cancellation'
    EXPIRATION = 'expiration'
    SERVICER_FAILURE = 'servicer failure'
    SERVICED_FAILURE = 'serviced failure'
    RECEPTION_FAILURE = 'reception failure'
    TRANSMISSION_FAILURE = 'transmission failure'


class ForeLink(object):
  """Accepts back-to-front tickets and emits front-to-back tickets."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def accept_back_to_front_ticket(self, ticket):
    """Accept a BackToFrontTicket.

    Args:
      ticket: Any BackToFrontTicket.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def join_rear_link(self, rear_link):
    """Mates this object with a peer with which it will exchange tickets."""
    raise NotImplementedError()


class RearLink(object):
  """Accepts front-to-back tickets and emits back-to-front tickets."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def accept_front_to_back_ticket(self, ticket):
    """Accepts a FrontToBackTicket.

    Args:
      ticket: Any FrontToBackTicket.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def join_fore_link(self, fore_link):
    """Mates this object with a peer with which it will exchange tickets."""
    raise NotImplementedError()


class FrontLink(Front, ForeLink):
  """Clientish objects that operate by sending and receiving tickets."""
  __metaclass__ = abc.ABCMeta


class BackLink(Back, RearLink):
  """Serverish objects that operate by sending and receiving tickets."""
  __metaclass__ = abc.ABCMeta



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""State used by both invocation-side and service-side code."""

import enum


@enum.unique
class HighWrite(enum.Enum):
  """The possible categories of high-level write state."""

  OPEN = 'OPEN'
  CLOSED = 'CLOSED'


class WriteState(object):
  """A description of the state of writing to an RPC.

  Attributes:
    low: A side-specific value describing the low-level state of writing.
    high: A HighWrite value describing the high-level state of writing.
    pending: A list of bytestrings for the RPC waiting to be written to the
      other side of the RPC.
  """

  def __init__(self, low, high, pending):
    self.low = low
    self.high = high
    self.pending = pending


class CommonRPCState(object):
  """A description of an RPC's state.

  Attributes:
    write: A WriteState describing the state of writing to the RPC.
    sequence_number: The lowest-unused sequence number for use in generating
      tickets locally describing the progress of the RPC.
    deserializer: The behavior to be used to deserialize payload bytestreams
      taken off the wire.
    serializer: The behavior to be used to serialize payloads to be sent on the
      wire.
  """

  def __init__(self, write, sequence_number, deserializer, serializer):
    self.write = write
    self.sequence_number = sequence_number
    self.deserializer = deserializer
    self.serializer = serializer
